# -*- coding: utf-8 -*-

import pytest
import socket

from wasp_general.network.web.proto import WWebSessionProto
from wasp_general.network.web.session import WWebSessionAdapter


class TestWWebSessionAdapter:

	test_socket = ('127.0.0.1', 5454)

	def test_adapter(self):
		assert(issubclass(WWebSessionAdapter, WWebSessionProto) is True)
		pytest.raises(NotImplementedError, WWebSessionAdapter.accepted_socket, None)

		server_socket = socket.socket()
		server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		server_socket.bind(TestWWebSessionAdapter.test_socket)
		server_socket.listen(1)

		client_socket = socket.socket()
		client_socket.connect(TestWWebSessionAdapter.test_socket)

		accepted_socket = server_socket.accept()[0]

		class Adapter(WWebSessionAdapter):

			def accepted_socket(self):
				return accepted_socket

			def protocol_version(self):
				return '0.9'

			def protocol(self):
				return 'http'

			def read_request(self):
				pass

			def write_response(self, request, response):
				pass

			def session_close(self):
				pass

		adapter = Adapter()
		assert(str(adapter.server_address().address()) == '127.0.0.1')
		assert(int(adapter.server_address().port()) == 5454)
		assert(str(adapter.client_address().address()) == '127.0.0.1')
		assert(int(adapter.client_address().port()) == accepted_socket.getpeername()[1])

		accepted_socket.close()
		client_socket.close()
		server_socket.close()



# -*- coding: utf-8 -*-
# wasp_general/task/registry.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_subclass, verify_type
from wasp_general.task.base import WTask


class WRegisteredTask(ABCMeta):
	""" Metaclass for task, that is stored in registry.  Derived class must redefine __registry__ property
	(and __registry_tag__ depends on registry storage - see
	:attr:`.WTaskRegistryStorage.__multiple_tasks_per_tag__`)
	"""

	__auto_registry__ = True
	""" Flag specifies whether this class will be registered automatically in registry or not
	"""

	__registry_tag__ = None
	""" Tag, that represent task. Depends on registry storage, independent task could be marked the same tag
	"""

	__registry__ = None
	""" Registry, that holds this task. Must be redefined in derived classes.
	"""

	def __init__(cls, name, bases, namespace):
		""" Construct new class. Derived class must redefine __registry__ property (and __registry_tag__
		depends on registry storage - see :attr:`.WTaskRegistryStorage.__multiple_tasks_per_tag__`)

		:param name: as name in type(cls, name, bases, namespace)
		:param bases: as bases in type(cls, name, bases, namespace)
		:param namespace: as namespace in type(cls, name, bases, namespace)
		"""
		ABCMeta.__init__(cls, name, bases, namespace)

		if cls.__auto_registry__ is not True:
			return

		if cls.__registry__ is None:
			raise ValueError('__registry__ must be defined')

		if issubclass(cls.__registry__, WTaskRegistry) is False:
			raise TypeError(
				"Property '__registry__' of tasks class has invalid type (must be"
				"WTaskRegistry or its subclass)"
			)

		if issubclass(cls, WTask) is False:
			raise TypeError('This class must inherit WTask class')

		cls.__registry__.add(cls)


class WTaskRegistryBase(metaclass=ABCMeta):
	""" Prototype for registry storage. Derived class must redefined following methods
	"""

	@abstractmethod
	def add(self, task):
		""" Add task to storage

		:param task: task to add (WTask class with WRegisteredTask metaclass)
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def remove(self, task):
		""" Remove task from storage

		:param task: task to remove (WTask class with WRegisteredTask metaclass)
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def clear(self):
		""" Remove every task from storage

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def tasks_by_tag(self, registry_tag):
		""" Get tasks from registry by its tag

		:param registry_tag: any hash-able object
		:return: Return task or list of tasks
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def tasks(self, task_cls=None):
		""" Return tasks that was added to this registry

		:param task_cls: if it is not None, then result will be consist of this subclass only (useful fo
		filtering tasks)

		:return: tuple of WRegisteredTask
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def count(self):
		""" Registered task count

		:return: int
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def tags(self):
		""" Return available registry tags

		:return: tuple of str
		"""
		raise NotImplementedError('This method is abstract')


class WTaskRegistryStorage(WTaskRegistryBase):
	""" Simple registry storage implementation
	"""

	__multiple_tasks_per_tag__ = True
	""" Flag that switches storing behaviour. If True - multiple task with the same tag are permitted,
	otherwise - task with the same tag is treated as error.
	"""

	def __init__(self):
		""" Construct new registry storage
		"""
		self.__registry = {}

	@verify_subclass(task_cls=WTask)
	@verify_type(task_cls=WRegisteredTask)
	def add(self, task_cls):
		""" Add task to this storage. Depends on :attr:`.WTaskRegistryStorage.__multiple_tasks_per_tag__`
		tasks with the same __registry_tag__ can be treated as error.

		:param task_cls: task to add
		:return: None
		"""

		registry_tag = task_cls.__registry_tag__
		if registry_tag not in self.__registry.keys():
			self.__registry[registry_tag] = [task_cls]
		elif self.__multiple_tasks_per_tag__ is True:
			self.__registry[registry_tag].append(task_cls)
		else:
			raise RuntimeError('Multiple tasks with same tag appended')

	@verify_subclass(task_cls=WTask)
	@verify_type(task_cls=WRegisteredTask)
	def remove(self, task_cls):
		""" Remove task from the storage. If task class are stored multiple times
		(if :attr:`.WTaskRegistryStorage.__multiple_tasks_per_tag__` is True) - removes all of them.

		:param task_cls: task to remove
		:return: None
		"""
		registry_tag = task_cls.__registry_tag__
		if registry_tag in self.__registry.keys():
			self.__registry[registry_tag] = \
				list(filter(lambda x: x != task_cls, self.__registry[registry_tag]))
			if len(self.__registry[registry_tag]) == 0:
				self.__registry.pop(registry_tag)

	def clear(self):
		""" Removes every task from storage

		:return: None
		"""
		self.__registry.clear()

	def tasks_by_tag(self, registry_tag):
		""" Get tasks from registry by its tag

		:param registry_tag: any hash-able object
		:return: Return task (if :attr:`.WTaskRegistryStorage.__multiple_tasks_per_tag__` is not True) or \
		list of tasks
		"""
		if registry_tag not in self.__registry.keys():
			return None
		tasks = self.__registry[registry_tag]
		return tasks if self.__multiple_tasks_per_tag__ is True else tasks[0]

	@verify_type(task_cls=(WRegisteredTask, None))
	def tasks(self, task_cls=None):
		""" :meth:`.WTaskRegistryBase.tasks` implementation
		"""
		result = []
		for tasks in self.__registry.values():
			result.extend(tasks)

		if task_cls is not None:
			result = filter(lambda x: issubclass(x, task_cls), result)
		return tuple(result)

	def tags(self):
		""" :meth:`.WTaskRegistryBase.tags` implementation
		"""
		return tuple(self.__registry.keys())

	def count(self):
		""" Registered task count

		:return: int
		"""
		result = 0
		for tasks in self.__registry.values():
			result += len(tasks)
		return result


class WTaskRegistry:
	""" Basic task registry. Derived classes must redefine __registry_storage__ property
	(see :attr:`.WTaskRegistry.__registry_storage__`)
	"""

	__registry_storage__ = None
	""" Storage that is used within this registry
	"""

	__skip_none_registry_tag__ = True
	""" Flag defines whether to register task which has __registry_tag__ set to None or not
	"""

	@classmethod
	def registry_storage(cls):
		""" Get registry storage

		:return: WTaskRegistryBase
		"""
		if cls.__registry_storage__ is None:
			raise ValueError('__registry_storage__ must be defined')
		if isinstance(cls.__registry_storage__, WTaskRegistryBase) is False:
			raise TypeError("Property '__registry_storage__' is invalid (must derived from WTaskRegistryBase)")

		return cls.__registry_storage__

	@classmethod
	@verify_subclass('paranoid', task_cls=WTask)
	@verify_type('paranoid', task_cls=WRegisteredTask)
	def add(cls, task_cls):
		""" Add task class to storage

		:param task_cls: task to add
		:return: None
		"""

		if task_cls.__registry_tag__ is None and cls.__skip_none_registry_tag__ is True:
			return

		cls.registry_storage().add(task_cls)

	@classmethod
	@verify_subclass('paranoid', task_cls=WTask)
	@verify_type('paranoid', task_cls=WRegisteredTask)
	def remove(cls, task_cls):
		""" Remove task class to storage

		:param task_cls: task to remove
		:return: None
		"""
		cls.registry_storage().remove(task_cls)

	@classmethod
	def clear(cls):
		""" Remove every task from storage

		:return: None
		"""
		cls.registry_storage().clear()



# -*- coding: utf-8 -*-

import pytest
import os
from tempfile import mktemp

from configparser import ConfigParser, NoOptionError, NoSectionError

from wasp_general.config import WConfig, WConfigSelection


@pytest.fixture
def tempfile(request):
	filename = mktemp('pytest-wasp-general-')

	def fin():
		if os.path.exists(filename):
			os.unlink(filename)
	request.addfinalizer(fin)
	return filename


class TestWConfig:

	def test_config(self, tempfile):
		conf = WConfig()
		assert(isinstance(conf, ConfigParser) is True)

		with pytest.raises(KeyError):
			print(conf['section1']['option1'])
		with pytest.raises(KeyError):
			print(conf['section1']['option2'])
		with pytest.raises(KeyError):
			print(conf['section1']['option3'])

		conf_parser = ConfigParser()
		conf_parser.add_section('section1')
		conf_parser.set('section1', 'option1', '1')

		conf.merge(conf_parser)
		assert(conf['section1']['option1'] == '1')
		with pytest.raises(KeyError):
			print(conf['section1']['option2'])
		with pytest.raises(KeyError):
			print(conf['section1']['option3'])

		with open(tempfile, 'w') as f:
			f.write('''
			[section1]
			option1 = 2
			option2 = foo, bar
			option3 =
			''')

		conf.merge(tempfile)
		assert (conf['section1']['option1'] == '2')
		assert (conf['section1']['option2'] == 'foo, bar')
		assert (conf['section1']['option3'] == '')

		assert(conf.split_option('section1', 'option2') == ['foo', 'bar'])
		assert(conf.split_option('section1', 'option3') == [])

	def test_merge(self):
		config1 = WConfig()
		config2 = WConfig()

		config1.add_section('section1.1')
		config1['section1.1']['option1'] = 'value1'
		config1['section1.1']['option2'] = '2'
		config1.add_section('section1.2')
		config1['section1.2']['option1'] = 'value1.2'
		config1['section1.2']['option2'] = '5'

		config2.add_section('section1.1')
		config2['section1.1']['option2'] = '7'
		config2['section1.1']['option3'] = '3'
		config2.add_section('section1.3')
		config2['section1.3']['option1'] = 'value2'
		config2.add_section('section1.4')
		config2['section1.4']['option'] = 'value'

		config1.merge_section(config2, 'section1.1')
		assert(config1['section1.1']['option1'] == 'value1')
		assert(config1['section1.1']['option2'] == '7')
		assert(config1['section1.2']['option1'] == 'value1.2')
		assert(config1['section1.2']['option2'] == '5')
		assert(config1.has_section('section1.4') is False)

		config1.merge_section(config2, 'section1.1', 'section1.3')

		assert(config1['section1.1']['option1'] == 'value2')
		assert(config1['section1.1']['option2'] == '7')
		assert(config1['section1.2']['option1'] == 'value1.2')
		assert(config1['section1.2']['option2'] == '5')
		assert(config1.has_section('section1.4') is False)

		config1.merge_section(config2, 'section1.4')
		assert(config1['section1.1']['option1'] == 'value2')
		assert(config1['section1.1']['option2'] == '7')
		assert(config1['section1.2']['option1'] == 'value1.2')
		assert(config1['section1.2']['option2'] == '5')
		assert(config1.has_section('section1.4') is True)

		pytest.raises(ValueError, config1.merge_section, config2, 'section1.2')


class TestWConfigSelection:

	def test(self):
		conf = WConfig()
		conf.add_section('section1')
		conf.add_section('section2')
		conf['section1']['option1'] = 'value1'
		conf['section1']['option1.sub-option1'] = '10'
		conf['section1']['option2'] = 'value2'
		conf['section2']['option2'] = 'true'

		pytest.raises(NoSectionError, conf.select_options, 'section0')

		conf_selection = conf.select_options('section1')
		assert(isinstance(conf_selection, WConfigSelection) is True)
		assert(conf_selection.config() == conf)
		assert(conf_selection.section() == 'section1')
		assert(conf_selection.option_prefix() == '')

		pytest.raises(NoOptionError, str, conf_selection)
		pytest.raises(NoOptionError, int, conf_selection)
		pytest.raises(NoOptionError, float, conf_selection)
		pytest.raises(NoOptionError, bool, conf_selection)

		conf_selection = conf_selection.select_options('option1')
		assert(isinstance(conf_selection, WConfigSelection) is True)
		assert(conf_selection.has_option() is True)
		assert(conf_selection['test_option'].has_option() is False)
		assert(str(conf_selection) == 'value1')
		pytest.raises(ValueError, int, conf_selection)
		pytest.raises(ValueError, float, conf_selection)
		pytest.raises(ValueError, bool, conf_selection)

		assert(int(conf_selection['.sub-option1']) == 10)
		assert(conf_selection.has_option('.sub-option1') is True)
		assert(conf_selection.has_option('.sub-option2') is False)
		assert(bool(conf.select_options('section2', 'option2')) is True)



# -*- coding: utf-8 -*-
# wasp_general/cli/curses_commands.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type
from wasp_general.command.command import WCommand
from wasp_general.command.result import WPlainCommandResult
from wasp_general.cli.curses import WCursesConsole


class WExitCommand(WCommand):

	@verify_type(console=WCursesConsole)
	def __init__(self, console):
		WCommand.__init__(self)
		self.__console = console

	def console(self):
		return self.__console

	@verify_type('paranoid', command_tokens=str)
	def match(self, *command_tokens, **command_env):
		return command_tokens == ('exit',) or command_tokens == ('quit',)

	@verify_type('paranoid', command_tokens=str)
	def _exec(self, *command_tokens, **command_env):
		self.__console.stop()
		return WPlainCommandResult('Exiting...')


class WEmptyCommand(WCommand):

	@verify_type('paranoid', command_tokens=str)
	def match(self, *command_tokens, **command_env):
		return len(command_tokens) == 0

	@verify_type('paranoid', command_tokens=str)
	def _exec(self, *command_tokens, **command_env):
		return WPlainCommandResult('')



# -*- coding: utf-8 -*-
# wasp_general/cli/formatter.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import time
import math

from wasp_general.verify import verify_type

from wasp_general.datetime import local_datetime


def local_datetime_formatter(dt):
	return '%s%s' % (local_datetime(dt=dt).isoformat(), time.strftime('%Z'))


@verify_type(none_value=(str, None))
def na_formatter(value, str_fn=None, none_value=None):
	fn = str_fn if str_fn is not None else str
	return fn(value) if value is not None else (none_value if none_value is not None else '(not available)')


@verify_type(size=int)
def data_size_formatter(size):
	suffixes = ['bytes', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB']
	suffix_index = int(math.floor(math.log2(size) / 10))
	if suffix_index >= len(suffixes):
		suffix_index = len(suffixes) - 1
	return '{:.2f} {}'.format(size / (1 << (suffix_index * 10)), suffixes[suffix_index])


class WConsoleTableFormatter:

	__default_delimiter__ = '*'

	@verify_type(table_headers=str)
	def __init__(self, *table_headers):
		self.__headers = table_headers
		self.__rows = []

		self.__cells_length = WConsoleTableFormatter.cells_length(*table_headers)

	@staticmethod
	def cells_length(*cells):
		return tuple([len(x) for x in cells])

	def add_row(self, *cells):
		self.__rows.append(cells)
		row_length = WConsoleTableFormatter.cells_length(*cells)

		min_cells, max_cells = row_length, self.__cells_length
		if len(cells) > len(max_cells):
			min_cells, max_cells = max_cells, min_cells

		result = [max(min_cells[i], max_cells[i]) for i in range(len(min_cells))]
		result.extend([max_cells[i] for i in range(len(min_cells), len(max_cells))])
		self.__cells_length = tuple(result)

	def format(self, delimiter=None):
		if delimiter is None:
			delimiter = self.__default_delimiter__

		cell_count = len(self.__cells_length)
		if cell_count == 0:
			raise RuntimeError('Empty table')

		separator_length = ((cell_count - 1) * 3) + 4
		for cell_length_iter in self.__cells_length:
			separator_length += cell_length_iter

		separator = (delimiter * separator_length) + '\n'
		left_border = '%s ' % delimiter
		int_border = ' %s ' % delimiter
		right_border = ' %s\n' % delimiter

		def render_row(*cells):
			row_result = ''
			for i in range(cell_count):
				cell_length = self.__cells_length[i]
				if i < len(cells):
					single_cell = cells[i]
					row_result += single_cell
					delta = (cell_length - len(single_cell))
				else:
					delta = cell_length
				row_result += ' ' * delta

				if i < (cell_count - 1):
					row_result += int_border

			return row_result

		result = separator
		result += left_border
		result += render_row(*self.__headers)
		result += right_border
		result += separator

		for row in self.__rows:
			result += left_border
			result += render_row(*row)
			result += right_border

		result += separator

		return result



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.web.proto import WWebResponseProto
from wasp_general.template import WTemplate, WTemplateText, WTemplateRenderer
from wasp_general.network.web.template_response import WWebTemplateResponse


class TestWWebTemplateResponse:

	class CorruptedTemplate(WTemplate):

		def template(self):
			return

	def test_response(self):
		response = WWebTemplateResponse(WTemplateText('code'))

		assert(isinstance(response, WWebTemplateResponse) is True)
		assert(isinstance(response, WTemplateRenderer) is True)
		assert(isinstance(response, WWebResponseProto) is True)
		assert(response.response_data() == 'code')



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.messenger.proto import WMessengerOnionProto, WMessengerOnionLayerProto

from wasp_general.network.messenger.onion import WMessengerOnion


class TestWMessengerOnion:

	class Layer(WMessengerOnionLayerProto):
		def process(self, envelope, session, **kwargs):
			pass

	def test(self):
		assert(isinstance(WMessengerOnion(), WMessengerOnion) is True)
		assert(isinstance(WMessengerOnion(), WMessengerOnionProto) is True)

		o = WMessengerOnion()
		assert(len(o.layers_names()) == 11)

		layer1 = TestWMessengerOnion.Layer('layer1')
		layer2 = TestWMessengerOnion.Layer('layer2')

		o = WMessengerOnion(layer1, layer2)
		layers = o.layers_names()
		layers.sort()
		assert(len(layers) == 13)
		assert('layer1' in layers)
		assert('layer2' in layers)

		pytest.raises(ValueError, o.add_layers, TestWMessengerOnion.Layer('layer1'))

		assert(o.layer('layer1') == layer1)
		assert(o.layer('layer2') == layer2)

		pytest.raises(RuntimeError, o.layer, 'layer3')



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.web.cookies import WHTTPCookie
from wasp_general.network.web.headers import WHTTPHeaders


class TestWHTTPHeaders:

	def test_headers(self):
		headers1 = WHTTPHeaders()
		assert(isinstance(headers1, WHTTPHeaders) is True)

		headers2 = WHTTPHeaders(header1='value1', **{'header-2': 'value-2'})
		assert(isinstance(headers2, WHTTPHeaders) is True)

		assert(headers1.headers() == tuple())
		headers_list = list(headers2.headers())
		headers_list.sort()
		assert(tuple(headers_list) == ('Header-2', 'Header1'))
		headers2 = headers2.switch_name_style('2')
		headers2.set_cookie_jar().add_cookie(WHTTPCookie('cookie', 'value'))
		headers_list = list(headers2.headers())
		headers_list.sort()
		assert(tuple(headers_list) == ('header-2', 'header1'))
		headers2 = headers2.switch_name_style('1.1')
		headers_list = list(headers2.headers())
		headers_list.sort()
		assert(tuple(headers_list) == ('Header-2', 'Header1'))
		headers2._WHTTPHeaders__normalization_mode = '3'
		pytest.raises(RuntimeError, headers2.normalize_name, 'header')
		headers2._WHTTPHeaders__normalization_mode = '1.1'

		headers1.add_headers('Header-Foo', 'bar')
		assert(headers1['header-foo'] == ('bar',))
		assert(headers1.get_headers('HeadEr-fOO') == ('bar',))
		headers1.add_headers('Header-Foo', 'zzz', 'yyy')
		assert(headers1['header-foo'] == ('bar', 'zzz', 'yyy'))
		headers1.replace_headers('header-foo', 'xxx', 'ccc')
		assert(headers1['header-foo'] == ('xxx', 'ccc'))
		headers1.remove_headers('header-Foo')
		assert(headers1.headers() == tuple())

		ro_headers = headers2.ro()
		assert(isinstance(ro_headers, WHTTPHeaders) is True)
		assert(ro_headers['header1'] == ('value1',))
		pytest.raises(RuntimeError, ro_headers.add_headers, 'header', 'value')
		pytest.raises(RuntimeError, ro_headers.remove_headers, 'header')
		pytest.raises(RuntimeError, ro_headers.replace_headers, 'header', '!')

		headers1.add_headers('content-type', 'foo/bar', 'text/plain')
		assert(headers1['content-type'] == ('foo/bar', 'text/plain'))
		assert(headers1.content_type() == 'foo/bar')
		assert(headers1.content_type('zzz') == 'zzz')
		assert(headers1.content_type() == 'zzz')

		headers1.add_headers('Cookie', 'client_cookie=val1; Path=/')
		assert(headers1.client_cookie_jar().cookies() == ('client_cookie',))
		assert(headers1.client_cookie_jar()['client_cookie'].value() == 'val1')
		assert(headers1.client_cookie_jar()['client_cookie'].attr('path') == '/')
		pytest.raises(RuntimeError, headers1.client_cookie_jar().remove_cookie, 'client_cookie')
		pytest.raises(RuntimeError, headers1.client_cookie_jar()['client_cookie'].value, 'aaa')

	def test_import(self):
		http_code = 'Header1: value1\r\n'
		http_code += 'Header-Foo: bar\r\n'
		http_code += 'Set-Cookie: cookie1=cookie_value1; Path=/\r\n'
		http_code += 'Cookie: cookie2=value2; Domain=zzz\r\n'

		headers = WHTTPHeaders.import_headers(http_code)
		assert(headers['Header1'] == ('value1',))
		assert(headers['Header-Foo'] == ('bar',))
		assert(headers.set_cookie_jar()['cookie1'].value() == 'cookie_value1')
		assert(headers.set_cookie_jar()['cookie1'].attr('Path') == '/')
		assert(headers.client_cookie_jar()['cookie2'].value() == 'value2')
		assert(headers.client_cookie_jar()['cookie2'].attr('Domain') == 'zzz')



# -*- coding: utf-8 -*-
# wasp_general/types/bytearray.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type, verify_value
from wasp_general.types.binarray import WBinArray


class WFixedSizeByteArray:
	""" This class represent fixed-length byte-array. Where byte is WBinArray. Bytes are ordered as big-endian.
	It means that the most significant byte has index 0.

	see also https://en.wikipedia.org/wiki/Endianness
	"""

	byte_size = 8
	""" Bits in byte. Derived classed can modify this without breaking the code. Derived class can do this,
	but why?!
	"""

	@verify_type(size=int, value=(WBinArray, int, bytes, None))
	@verify_value(size=lambda x: x >= 0)
	def __init__(self, size=0, value=None):
		""" Construct new array.

		:param size: count of bytes
		:param value: value with which this sequence must be initialized (default 0)
		"""
		self.__array = []
		self.__size = size

		for i in range(self.__size):
			self.__array.append(WBinArray(0, self.__class__.byte_size))

		if value is not None:
			if isinstance(value, (WBinArray, int)) is True:
				value = WBinArray(int(value), self.__size * self.__class__.byte_size)
				value = value.split(self.__class__.byte_size)
			else:
				# isinstance(value, bytes)
				if self.__size < len(value):
					raise OverflowError('Value is out of bound')

			for i in range(len(value)):
				self.__array[i] = value[i]

	def bin_value(self):
		""" Return this sequence as single big WBinArray

		:return: WBinArray
		"""
		return WBinArray.join(*self.__array)

	def bin_array(self):
		""" Return this sequence as list of bytes (WBinArray)

		:return: list of WBinArray
		"""
		return self.__array

	def resize(self, size):
		""" Grow this array to specified length. This array can't be shrinked

		:param size: new length
		:return: None
		"""
		if size < len(self):
			raise ValueError("Value is out of bound. Array can't be shrinked")
		current_size = self.__size
		for i in range(size - current_size):
			self.__array.append(WBinArray(0, self.__class__.byte_size))
		self.__size = size

	def __len__(self):
		""" Return count of bytes

		:return: int
		"""
		return self.__size

	def __str__(self):
		""" Convert to string

		:return: str
		"""
		return str([str(x) for x in self.__array])

	def __getitem__(self, item):
		""" Return byte (WBinArray) at specified index

		:param item: item index
		:return: WBinArray
		"""
		return self.__array[item]

	@verify_type('paranoid', value=(WBinArray, int))
	@verify_type(key=int)
	def __setitem__(self, key, value):
		""" Set value for the given index. Specified value must resign within byte capability (must be non
		negative and be less then 2^<byte_size>).

		:param key: item index
		:param value: value to set
		:return: None
		"""
		if key < 0 or key >= len(self):
			raise IndexError('Index out of range')

		self.__array[key] = WBinArray(value, self.__class__.byte_size)

	def __bytes__(self):
		""" Convert to bytes

		:return: bytes
		"""
		return bytes([int(x) for x in self.__array])

	def swipe(self):
		""" Mirror current array value in reverse. Bytes that had greater index will have lesser index, and
		vice-versa. This method doesn't change this array. It creates a new one and return it as a result.

		:return: WFixedSizeByteArray
		"""
		result = WFixedSizeByteArray(len(self))
		for i in range(len(self)):
			result[len(self) - i - 1] = self[i]
		return result



# -*- coding: utf-8 -*-

import io
import os
import pytest
from threading import Thread, Lock
from zmq.eventloop.ioloop import IOLoop
from tempfile import mktemp


from wasp_general.network.transport import WNetworkNativeTransportProto
from wasp_general.network.beacon.beacon import WNetworkServerBeacon, WNetworkClientBeacon, WNetworkBeaconCallback
from wasp_general.network.beacon.messenger import WBeaconMessenger
from wasp_general.network.beacon.transport import WBroadcastBeaconTransport, WMulticastBeaconTransport

from wasp_general.config import WConfig
from wasp_general.network.primitives import WIPV4SocketInfo


def test_abstract():
	pytest.raises(TypeError, WNetworkBeaconCallback)
	description = WNetworkBeaconCallback.WDataDescription.request
	pytest.raises(NotImplementedError, WNetworkBeaconCallback.__call__, None, b'', WIPV4SocketInfo(), description)


class FakeBeaconTransport(WNetworkNativeTransportProto):

	__global_lock__ = Lock()

	class FakeSocket:

		def __init__(self, ro_filename, wo_filename):
			self.__ro_filename = ro_filename
			self.__wo_filename = wo_filename
			self.__file_ro = None
			self.__file_wo = None

		def open(self):
			if self.__file_ro is None:
				self.__file_ro = open(self.__ro_filename, 'r')
			if self.__file_wo is None:
				self.__file_wo = open(self.__wo_filename, 'w')

		def fileno(self):
			return self.__file_ro.fileno()

		def recvfrom(self, bytes_count):
			FakeBeaconTransport.__global_lock__.acquire()
			self.__file_ro.seek(0, io.SEEK_SET)
			data = self.__file_ro.read(bytes_count).encode()
			truncate_file = open(self.__ro_filename, 'w')
			truncate_file.seek(0, io.SEEK_SET)
			truncate_file.truncate()
			FakeBeaconTransport.__global_lock__.release()
			return data, ('fake-host', 1)

		def sendto(self, data, address):
			FakeBeaconTransport.__global_lock__.acquire()
			self.__file_wo.seek(0, io.SEEK_SET)
			self.__file_wo.write(data.decode())
			self.__file_wo.flush()
			FakeBeaconTransport.__global_lock__.release()

	def __init__(self, file_suffix):
		WNetworkNativeTransportProto.__init__(self)
		self.__file_suffix = file_suffix
		self.__server_filename = None
		self.__client_filename = None
		self.__server_file = None
		self.__client_file = None

	def tempfile(self):
		filename = mktemp(self.__file_suffix)
		with open(filename, 'w'):
			pass
		return filename

	def create_files(self):
		force_recreation = False
		if self.__server_filename is None:
			self.__server_filename = self.tempfile()
			force_recreation = True

		if self.__client_filename is None:
			self.__client_filename = self.tempfile()
			force_recreation = True

		if self.__server_file is None or force_recreation is True:
			self.__server_file = FakeBeaconTransport.FakeSocket(
				self.__server_filename, self.__client_filename
			)
		if self.__client_file is None or force_recreation is True:
			self.__client_file = FakeBeaconTransport.FakeSocket(
				self.__client_filename, self.__server_filename
			)

	def delete_files(self):
		if self.__server_file is not None:
			self.__server_file = None

		if self.__client_file is not None:
			self.__client_file = None

		if self.__server_filename is not None:
			os.unlink(self.__server_filename)
			self.__server_filename = None

		if self.__client_filename is not None:
			os.unlink(self.__client_filename)
			self.__client_filename = None

	def server_socket(self, beacon_config):
		self.__server_file.open()
		return self.__server_file

	def close_server_socket(self, beacon_config, close_fd=True):
		pass

	def client_socket(self, beacon_config):
		self.__client_file.open()
		return self.__client_file

	def close_client_socket(self, beacon_config, close_fd=True):
		pass

	def target_socket(self, beacon_config):
		return WIPV4SocketInfo('fake-host', 1)


class TestWBeacon:

	def test_beacon(self):

		class TestCallback(WNetworkBeaconCallback):
			result = []

			def __call__(self, data, source, description):
				TestCallback.result.append(str(description).encode() + b':' + data)

		fake_transport = FakeBeaconTransport('-pytest-wasp-general')
		fake_transport.create_files()
		messenger = WBeaconMessenger()
		config = WConfig()
		config.add_section('wasp-general::network::beacon')
		config['wasp-general::network::beacon']['lookup_timeout'] = '1'

		server_beacon = WNetworkServerBeacon(messenger=messenger, transport=fake_transport, callback=TestCallback())
		assert(server_beacon.handler().io_handler().process_any() is False)
		client_beacon = WNetworkClientBeacon(
			config=config, messenger=messenger, transport=fake_transport, callback=TestCallback()
		)

		assert(isinstance(server_beacon.loop(), IOLoop) is True)
		assert(isinstance(client_beacon.loop(), IOLoop) is True)

		server_thread = Thread(target=server_beacon.start)
		server_thread.start()
		client_beacon.start()
		client_beacon.stop()
		server_beacon.stop()
		server_thread.join()

		assert(TestCallback.result == [b'WDataDescription.request:HELLO', b'WDataDescription.response:HELLO'])

		assert(isinstance(WNetworkServerBeacon().handler().transport(), WBroadcastBeaconTransport) is True)
		config['wasp-general::network::beacon']['transport'] = 'multicast'
		assert(isinstance(WNetworkServerBeacon(config=config).handler().transport(), WMulticastBeaconTransport) is True)
		config['wasp-general::network::beacon']['transport'] = 'foo'
		pytest.raises(ValueError, WNetworkServerBeacon, config=config)
		config['wasp-general::network::beacon']['transport'] = 'unicast_udp'
		pytest.raises(NotImplementedError, WNetworkServerBeacon, config=config)
		config['wasp-general::network::beacon']['transport'] = 'unicast_tcp'
		pytest.raises(NotImplementedError, WNetworkServerBeacon, config=config)

		silent_messenger = WBeaconMessenger()
		silent_messenger.has_response = lambda c, r, a: False

		server_beacon = WNetworkServerBeacon(
			messenger=silent_messenger, transport=fake_transport, callback=TestCallback(),
			process_any=True
		)
		assert(server_beacon.handler().io_handler().process_any() is True)
		TestCallback.result = []
		server_thread = Thread(target=server_beacon.start)
		server_thread.start()
		client_beacon.start()
		client_beacon.stop()
		server_beacon.stop()
		server_thread.join()

		assert (TestCallback.result == [b'WDataDescription.invalid_request:HELLO'])

		fake_transport.delete_files()



# -*- coding: utf-8 -*-
# wasp_general/network/beacon/task.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.task.thread import WThreadTask

from wasp_general.network.beacon.beacon import WNetworkServerBeacon


class WNetworkBeaconTask(WNetworkServerBeacon, WThreadTask):
	""" Convenient way to start WNetworkBeacon server side

	see :class:`.WNetworkServerBeacon`
	"""

	def __init__(self, config=None, config_section=None, thread_name=None, messenger=None):
		""" Create new threaded task

		:param config: same as config in :meth:`.WNetworkServerBeacon.__init__` method
		:param config_section: same as config_section in :meth:`.WNetworkServerBeacon.__init__` method
		:param thread_name: same as thread_name in :meth:`.WThreadTask.__init__` method
		:param messenger: same as messenger in :meth:`.WNetworkServerBeacon.__init__` method
		"""
		WNetworkServerBeacon.__init__(self, config=config, config_section=config_section, messenger=messenger)
		WThreadTask.__init__(self, thread_name=thread_name)

	def start(self):
		WThreadTask.start(self)

	def stop(self):
		WThreadTask.stop(self)

	def thread_started(self):
		WNetworkServerBeacon.start(self)

	def thread_stopped(self):
		WNetworkServerBeacon.stop(self)



# -*- coding: utf-8 -*-

import pytest
import time

from wasp_general.task.thread import WThreadTask
from wasp_general.task.scheduler.proto import WScheduleTask, WScheduleRecord
from wasp_general.task.thread_tracker import WThreadTrackerInfoStorageProto, WThreadTracker, WSimpleTrackerStorage
from wasp_general.task.thread_tracker import WScheduleRecordTracker, WTrackerEvents


def test_abstract():
	pytest.raises(TypeError, WThreadTrackerInfoStorageProto)
	pytest.raises(
		NotImplementedError,
		WThreadTrackerInfoStorageProto.register_start,
		None, TestWThreadTracker.Task()
	)
	pytest.raises(
		NotImplementedError,
		WThreadTrackerInfoStorageProto.register_stop,
		None, TestWThreadTracker.Task()
	)
	pytest.raises(
		NotImplementedError,
		WThreadTrackerInfoStorageProto.register_termination,
		None, TestWThreadTracker.Task()
	)
	pytest.raises(
		NotImplementedError,
		WThreadTrackerInfoStorageProto.register_exception,
		None, TestWThreadTracker.Task(), ValueError('!'), '!'
	)
	pytest.raises(
		NotImplementedError,
		WThreadTrackerInfoStorageProto.register_wait,
		None, TestWThreadTracker.Task()
	)
	pytest.raises(
		NotImplementedError,
		WThreadTrackerInfoStorageProto.register_drop,
		None, TestWThreadTracker.Task()
	)

	pytest.raises(TypeError, WThreadTracker)


class TestWThreadTracker:

	class Task(WThreadTracker):

		def thread_started(self):
			pass

		def long_task(self):
			while self.stop_event().is_set() is False:
				time.sleep(0.01)

		@staticmethod
		def exc_task(*args, **kwargs):
			raise ValueError('!')

	class ScheduleTask(Task, WScheduleTask):

		def __init__(self, tracker_storage=None):
			TestWThreadTracker.Task.__init__(self, tracker_storage=tracker_storage)
			WScheduleTask.__init__(self)

	class Storage(WThreadTrackerInfoStorageProto):

		def __init__(self):
			self.start_count = 0
			self.stop_count = 0
			self.termination_count = 0
			self.exception_count = 0
			self.wait_count = 0
			self.drop_count = 0

		def register_start(self, task, event_details=None):
			self.start_count += 1

		def register_stop(self, task, event_details=None):
			self.stop_count += 1

		def register_termination(self, task, event_details=None):
			self.termination_count += 1

		def register_exception(self, task, raised_exception, exception_details, event_details=None):
			self.exception_count += 1

		def register_wait(self, task, event_details=None):
			self.wait_count += 1

		def register_drop(self, task, event_details=None):
			self.drop_count += 1

		def status(self):
			return (
				self.start_count, self.stop_count, self.termination_count, self.exception_count,
				self.wait_count, self.drop_count
			)

	def test(self):
		task = TestWThreadTracker.Task()
		assert(isinstance(task, TestWThreadTracker.Task) is True)
		assert(isinstance(task, WThreadTracker) is True)
		assert(isinstance(task, WThreadTask) is True)
		assert(task.tracker_storage() is None)
		assert(task.event_details(WTrackerEvents.start) is None)

		assert(task.track_stop() is True)
		assert(task.track_termination() is True)
		assert(task.track_exception() is True)

		task.start()
		task.ready_event().wait()
		task.stop()

		storage = TestWThreadTracker.Storage()
		assert(storage.status() == (0, 0, 0, 0, 0, 0))
		task = TestWThreadTracker.Task(tracker_storage=storage)
		assert(storage.status() == (0, 0, 0, 0, 0, 0))
		task.start()
		assert(storage.status() == (1, 0, 0, 0, 0, 0))
		task.ready_event().wait()
		task.stop()
		assert(storage.status() == (1, 1, 0, 0, 0, 0))

		task.thread_started = task.long_task
		assert(storage.status() == (1, 1, 0, 0, 0, 0))
		task.start()
		assert(storage.status() == (2, 1, 0, 0, 0, 0))
		task.start_event().wait()
		assert(storage.status() == (2, 1, 0, 0, 0, 0))
		task.stop()
		assert(storage.status() == (2, 1, 1, 0, 0, 0))

		task.thread_started = TestWThreadTracker.Task.exc_task
		assert(storage.status() == (2, 1, 1, 0, 0, 0))
		task.start()
		task.start_event().wait()
		task.stop()
		assert(storage.status() == (3, 1, 1, 1, 0, 0))

		storage.register_exception = TestWThreadTracker.Task.exc_task
		assert(storage.status() == (3, 1, 1, 1, 0, 0))
		task.start()
		task.start_event().wait()
		task.stop()
		assert(storage.status() == (4, 1, 1, 1, 0, 0))

		task = TestWThreadTracker.Task(
			tracker_storage=storage, track_start=False, track_stop=False, track_termination=False,
			track_exception=False
		)

		assert(task.track_stop() is False)
		assert(task.track_termination() is False)
		assert(task.track_exception() is False)

		task.start()
		task.ready_event().wait()
		task.stop()

		task.thread_started = task.long_task
		task.start()
		task.stop()

		task.thread_started = TestWThreadTracker.Task.exc_task
		task.start()
		task.stop()

		assert(storage.status() == (4, 1, 1, 1, 0, 0))

		storage.register_stop = TestWThreadTracker.Task.exc_task
		task = TestWThreadTracker.Task(tracker_storage=storage)
		task.start()
		task.ready_event().wait()
		task.stop()


class TestWSimpleTrackerStorage:

	def test(self):

		pytest.raises(TypeError, WSimpleTrackerStorage.Record, 1, TestWThreadTracker.Task())

		storage = WSimpleTrackerStorage()
		assert(isinstance(storage, WSimpleTrackerStorage) is True)
		assert(isinstance(storage, WThreadTrackerInfoStorageProto) is True)
		assert(storage.record_limit() is None)
		assert([] == [x for x in storage])

		task1 = TestWThreadTracker.Task(storage)
		task1.start()

		result = [x for x in storage]
		assert(len(result) == 1)
		assert(result[0].record_type == WTrackerEvents.start)
		assert(result[0].thread_task == task1)
		assert(result[0].event_details is None)

		task1.stop()

		result = [x for x in storage]
		assert(len(result) == 2)
		assert(result[0].record_type == WTrackerEvents.stop)
		assert(result[0].thread_task == task1)
		assert(result[0].event_details is None)
		assert(result[1].thread_task == task1)
		assert(result[1].record_type == WTrackerEvents.start)

		task2 = TestWThreadTracker.Task(storage)
		task2.event_details = lambda x: '!!!'
		task2.start()
		task2.stop()

		result = [x for x in storage]
		assert(len(result) == 4)
		assert(result[0].record_type == WTrackerEvents.stop)
		assert(result[0].thread_task == task2)
		assert(result[0].event_details == '!!!')
		assert(result[1].thread_task == task2)
		assert(result[2].thread_task == task1)
		assert(result[3].thread_task == task1)

		task3 = TestWThreadTracker.Task(storage)
		task3.thread_started = task3.long_task
		task3.start()
		task3.start_event().wait()
		task3.stop()

		result = [x for x in storage]
		assert(len(result) == 6)
		assert(result[0].record_type == WTrackerEvents.termination)
		assert(result[0].thread_task == task3)
		assert(result[0].event_details is None)
		assert(result[1].thread_task == task3)
		assert(result[2].thread_task == task2)
		assert(result[3].thread_task == task2)
		assert(result[4].thread_task == task1)
		assert(result[5].thread_task == task1)

		task4 = TestWThreadTracker.Task(storage)
		task4.thread_started = TestWThreadTracker.Task.exc_task
		task4.start()
		task4.start_event().wait()
		task4.stop()

		result = [x for x in storage]
		assert(len(result) == 8)
		assert(result[0].record_type == WTrackerEvents.exception)
		assert(result[0].thread_task == task4)
		assert(result[0].event_details is None)
		assert(result[1].thread_task == task4)
		assert(result[2].thread_task == task3)
		assert(result[3].thread_task == task3)
		assert(result[4].thread_task == task2)
		assert(result[5].thread_task == task2)
		assert(result[6].thread_task == task1)
		assert(result[7].thread_task == task1)

		task5 = TestWThreadTracker.ScheduleTask(storage)
		record1 = WScheduleRecordTracker(task5)
		record1.task_postponed()
		result = [x for x in storage]
		assert(len(result) == 9)
		assert(result[0].record_type == WTrackerEvents.wait)
		assert(result[0].thread_task == task5)
		assert(result[0].event_details is None)
		assert(result[1].thread_task == task4)
		assert(result[2].thread_task == task4)
		assert(result[3].thread_task == task3)
		assert(result[4].thread_task == task3)
		assert(result[5].thread_task == task2)
		assert(result[6].thread_task == task2)
		assert(result[7].thread_task == task1)
		assert(result[8].thread_task == task1)

		task6 = TestWThreadTracker.ScheduleTask(storage)
		record2 = WScheduleRecordTracker(task6)
		record2.task_dropped()
		result = [x for x in storage]
		assert(len(result) == 10)
		assert(result[0].record_type == WTrackerEvents.drop)
		assert(result[0].thread_task == task6)
		assert(result[0].event_details is None)
		assert(result[1].thread_task == task5)
		assert(result[2].thread_task == task4)
		assert(result[3].thread_task == task4)
		assert(result[4].thread_task == task3)
		assert(result[5].thread_task == task3)
		assert(result[6].thread_task == task2)
		assert(result[7].thread_task == task2)
		assert(result[8].thread_task == task1)
		assert(result[9].thread_task == task1)

		assert(storage.last_record(TestWThreadTracker.ScheduleTask(storage).uid()) is None)
		assert(isinstance(storage.last_record(task5.uid()), WSimpleTrackerStorage.Record) is True)
		assert(storage.last_record(task5.uid()).thread_task == task5)
		assert(storage.last_record(task5.uid(), WTrackerEvents.drop) is None)
		assert(storage.last_record(task5.uid(), WTrackerEvents.drop, WTrackerEvents.wait).thread_task == task5)

		storage = WSimpleTrackerStorage(records_limit=4)
		task1 = TestWThreadTracker.Task(storage)
		task1.start()
		task1.stop()

		task2 = TestWThreadTracker.Task(storage)
		task2.event_details = lambda x: '!!!'
		task2.start()
		task2.stop()

		task3 = TestWThreadTracker.Task(storage)
		task3.thread_started = task3.long_task
		task3.start()
		task3.start_event().wait()
		task3.stop()

		task4 = TestWThreadTracker.Task(storage)
		task4.thread_started = TestWThreadTracker.Task.exc_task
		task4.start()
		task4.start_event().wait()
		task4.stop()

		result = [x for x in storage]

		assert(len(result) == 4)
		assert(result[0].record_type == WTrackerEvents.exception)
		assert(result[0].thread_task == task4)
		assert(result[0].event_details is None)
		assert(result[1].thread_task == task4)
		assert(result[2].thread_task == task3)
		assert(result[3].thread_task == task3)

		storage = WSimpleTrackerStorage(
			record_start=False, record_stop=False, record_termination=False, record_exception=False,
			record_wait=False, record_drop=False
		)
		task1 = TestWThreadTracker.Task(storage)
		task1.start()
		task1.stop()

		task2 = TestWThreadTracker.Task(storage)
		task2.event_details = lambda x: '!!!'
		task2.start()
		task2.stop()

		task3 = TestWThreadTracker.Task(storage)
		task3.thread_started = task3.long_task
		task3.start()
		task3.start_event().wait()
		task3.stop()

		task4 = TestWThreadTracker.Task(storage)
		task4.thread_started = TestWThreadTracker.Task.exc_task
		task4.start()
		task4.start_event().wait()
		task4.stop()

		task5 = TestWThreadTracker.ScheduleTask(storage)
		record1 = WScheduleRecordTracker(task5)
		record1.task_postponed()

		task6 = TestWThreadTracker.ScheduleTask(storage)
		record2 = WScheduleRecordTracker(task6)
		record2.task_dropped()

		result = [x for x in storage]
		assert(len(result) == 0)

		pytest.raises(TypeError, storage._WSimpleTrackerStorage__store_record, 1)


class TestWScheduleRecordTracker:

	def test(self):
		storage = TestWThreadTracker.Storage()
		assert(storage.status() == (0, 0, 0, 0, 0, 0))

		task = TestWThreadTracker.ScheduleTask(tracker_storage=storage)
		record1 = WScheduleRecordTracker(task)
		assert(isinstance(record1, WScheduleRecordTracker))
		assert(isinstance(record1, WScheduleRecord))

		record1.task_postponed()
		assert (storage.status() == (0, 0, 0, 0, 1, 0))

		record1.task_dropped()
		assert (storage.status() == (0, 0, 0, 0, 1, 1))

		record2 = WScheduleRecordTracker(task, track_drop=False, track_wait=False)
		record2.task_postponed()
		record2.task_dropped()
		assert (storage.status() == (0, 0, 0, 0, 1, 1))



# -*- coding: utf-8 -*-

import datetime
import time

from wasp_general.datetime import local_tz, utc_datetime, local_datetime


def test_datetime():
	now_dt = utc_datetime()
	assert(now_dt.timetz().tzinfo == datetime.timezone.utc)

	test_dt = datetime.datetime(1970, 9, 8, 7, 0)
	utc_test_dt = utc_datetime(test_dt)
	assert(utc_test_dt.timetz().tzinfo == datetime.timezone.utc)
	assert(utc_test_dt.utcoffset().seconds == 0)

	compare_utc_test_dt = test_dt + datetime.timedelta(seconds=time.timezone)
	assert(compare_utc_test_dt.year == utc_test_dt.year)
	assert(compare_utc_test_dt.month == utc_test_dt.month)
	assert(compare_utc_test_dt.day == utc_test_dt.day)
	assert(compare_utc_test_dt.hour == utc_test_dt.hour)
	assert(compare_utc_test_dt.minute == utc_test_dt.minute)
	assert(compare_utc_test_dt.second == utc_test_dt.second)

	utc_test_dt = utc_datetime(test_dt, local_value=False)
	assert(utc_test_dt.timetz().tzinfo == datetime.timezone.utc)
	assert(utc_test_dt.year == 1970)
	assert(utc_test_dt.month == 9)
	assert(utc_test_dt.day == 8)
	assert(utc_test_dt.hour == 7)
	assert(utc_test_dt.minute == 0)
	assert(utc_test_dt.second == 0)

	now_dt = local_datetime()
	assert(now_dt.utcoffset().seconds == (time.timezone * -1))

	local_test_dt = local_datetime(test_dt)
	assert(local_test_dt.timetz().utcoffset().seconds == (time.timezone * -1))
	compare_local_test_dt = test_dt + datetime.timedelta(seconds=(time.timezone * -1))
	assert(compare_local_test_dt.year == local_test_dt.year)
	assert(compare_local_test_dt.month == local_test_dt.month)
	assert(compare_local_test_dt.day == local_test_dt.day)
	assert(compare_local_test_dt.hour == local_test_dt.hour)
	assert(compare_local_test_dt.minute == local_test_dt.minute)
	assert(compare_local_test_dt.second == local_test_dt.second)

	local_test_dt = local_datetime(test_dt, utc_value=False)
	assert(local_test_dt.timetz().utcoffset().seconds == (time.timezone * -1))
	assert(local_test_dt.year == 1970)
	assert(local_test_dt.month == 9)
	assert(local_test_dt.day == 8)
	assert(local_test_dt.hour == 7)
	assert(local_test_dt.minute == 0)
	assert(local_test_dt.second == 0)



# -*- coding: utf-8 -*-

import pytest
import socket

from wasp_general.config import WConfig

from wasp_general.network.transport import WNetworkNativeTransportProto, WNetworkNativeTransport
from wasp_general.network.transport import WNetworkNativeTransportSocketConfig, WUDPNetworkNativeTransport
from wasp_general.network.transport import WMulticastNetworkTransport, WTCPNetworkNativeTransport
from wasp_general.network.transport import WBroadcastNetworkTransport
from wasp_general.network.primitives import WIPV4SocketInfo


def test_abstract():
	pytest.raises(TypeError, WNetworkNativeTransportProto)
	pytest.raises(NotImplementedError, WNetworkNativeTransportProto.server_socket, None, WConfig())
	pytest.raises(NotImplementedError, WNetworkNativeTransportProto.close_server_socket, None, WConfig())
	pytest.raises(NotImplementedError, WNetworkNativeTransportProto.client_socket, None, WConfig())
	pytest.raises(NotImplementedError, WNetworkNativeTransportProto.close_client_socket, None, WConfig())
	pytest.raises(NotImplementedError, WNetworkNativeTransportProto.target_socket, None, WConfig())
	pytest.raises(NotImplementedError, WNetworkNativeTransportProto.bind_socket, None, WConfig())

	pytest.raises(TypeError, WNetworkNativeTransport)
	pytest.raises(NotImplementedError, WNetworkNativeTransport._create_socket, None)


class TestWNetworkNativeTransportSocketConfig:

	def test_config(self):
		c = WNetworkNativeTransportSocketConfig('section1', 'option1', 'option2')
		assert(c.section == 'section1')
		assert(c.address_option == 'option1')
		assert(c.port_option == 'option2')


class TestNetworkTransport:

	class NativeTransport(WNetworkNativeTransport):

		def _create_socket(self):
			return socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

	def test_bind_socket(self):
		config = WConfig()
		config.add_section('section')
		config['section']['server_address'] = ''
		config['section']['server_port'] = ''
		config['section']['client_address'] = ''
		config['section']['client_port'] = ''

		assert(issubclass(WNetworkNativeTransport, WNetworkNativeTransportProto) is True)

		t = TestNetworkTransport.NativeTransport(
			WNetworkNativeTransportSocketConfig('section', 'client_address', 'client_port'),
			WNetworkNativeTransportSocketConfig('section', 'server_address', 'server_port')
		)

		pytest.raises(ValueError, t.bind_socket, config)
		config['section']['server_address'] = 'site1.org'
		pytest.raises(ValueError, t.bind_socket, config)
		config['section']['server_address'] = ''
		config['section']['server_port'] = '70'
		assert(isinstance(t.bind_socket(config), WIPV4SocketInfo) is True)
		assert(t.bind_socket(config).pair() == ('', 70))

		config['section']['server_address'] = 'site1.org'
		assert(t.bind_socket(config).pair() == ('site1.org', 70))

		config['section']['server_address'] = '192.192.192.192'
		assert (t.bind_socket(config).pair() == ('192.192.192.192', 70))

		pytest.raises(ValueError, t.target_socket, config)
		config['section']['client_address'] = 'site1.org'
		pytest.raises(ValueError, t.target_socket, config)
		config['section']['client_address'] = ''
		config['section']['client_port'] = '9999'
		pytest.raises(ValueError, t.target_socket, config)

		config['section']['client_address'] = 'site1.org'
		assert(isinstance(t.target_socket(config), WIPV4SocketInfo) is True)
		assert(t.target_socket(config).pair() == ('site1.org', 9999))

		config['section']['client_address'] = '4.4.4.4'
		assert(isinstance(t.target_socket(config), WIPV4SocketInfo) is True)
		assert(t.target_socket(config).pair() == ('4.4.4.4', 9999))

	def test_socket(self):

		config = WConfig()
		config.add_section('section')
		config['section']['server_address'] = ''
		config['section']['server_port'] = '20202'
		config['section']['client_address'] = '127.0.0.1'
		config['section']['client_port'] = '20202'

		t = TestNetworkTransport.NativeTransport(
			WNetworkNativeTransportSocketConfig('section', 'client_address', 'client_port'),
			WNetworkNativeTransportSocketConfig('section', 'server_address', 'server_port')
		)
		s1 = t.server_socket(config)
		assert(isinstance(s1, socket.socket) is True)
		assert(s1.getsockname() == ('0.0.0.0', 20202))
		assert(s1 == t.server_socket(config))
		t.close_server_socket(config)

		config['section']['server_address'] = '127.0.0.1'
		s2 = t.server_socket(config)
		assert(s2.getsockname() == ('127.0.0.1', 20202))
		assert(s2 != s1)
		t.close_server_socket(config)

		s1 = t.client_socket(config)
		assert(isinstance(s1, socket.socket) is True)
		assert(s1 == t.client_socket(config))
		t.close_client_socket(config)

		s2 = t.client_socket(config)
		assert(s2 != s1)
		t.close_client_socket(config)


class TestWUDPNetworkTransport:

	def test_socket(self):

		config = WConfig()
		config.add_section('section')
		config['section']['server_address'] = ''
		config['section']['server_port'] = '20202'
		config['section']['client_address'] = '127.0.0.1'
		config['section']['client_port'] = '20202'

		t = WUDPNetworkNativeTransport(
			WNetworkNativeTransportSocketConfig('section', 'client_address', 'client_port'),
			WNetworkNativeTransportSocketConfig('section', 'server_address', 'server_port')
		)
		s = t._create_socket()
		assert(isinstance(s, socket.socket) is True)
		assert(s.type == socket.SOCK_DGRAM)


class TestWTCPNetworkTransport:

	def test_socket(self):

		config = WConfig()
		config.add_section('section')
		config['section']['server_address'] = ''
		config['section']['server_port'] = '20202'
		config['section']['client_address'] = '127.0.0.1'
		config['section']['client_port'] = '20202'

		t = WTCPNetworkNativeTransport(
			WNetworkNativeTransportSocketConfig('section', 'client_address', 'client_port'),
			WNetworkNativeTransportSocketConfig('section', 'server_address', 'server_port')
		)
		s = t._create_socket()
		assert(isinstance(s, socket.socket) is True)
		assert(s.type == socket.SOCK_STREAM)


class TestWBroadcastNetworkTransport:

	def test_target_socket(self):
		config = WConfig()
		config.add_section('section')
		config['section']['server_address'] = ''
		config['section']['server_port'] = ''
		config['section']['client_address'] = ''
		config['section']['client_port'] = ''

		t = WBroadcastNetworkTransport(
			WNetworkNativeTransportSocketConfig('section', 'client_address', 'client_port'),
			WNetworkNativeTransportSocketConfig('section', 'server_address', 'server_port')
		)
		pytest.raises(ValueError, t.target_socket, config)
		config['section']['client_address'] = 'site1.org'
		pytest.raises(ValueError, t.target_socket, config)
		config['section']['client_address'] = ''
		config['section']['client_port'] = '9999'
		pytest.raises(ValueError, t.target_socket, config)

		config['section']['client_address'] = 'site1.org'
		pytest.raises(ValueError, t.target_socket, config)

		config['section']['client_address'] = '4.4.4.4'
		assert(isinstance(t.target_socket(config), WIPV4SocketInfo) is True)
		assert(t.target_socket(config).pair() == ('4.4.4.4', 9999))

	def test_client_socket(self):
		config = WConfig()
		config.add_section('section')
		config['section']['server_address'] = ''
		config['section']['server_port'] = ''
		config['section']['client_address'] = '4.4.4.4'
		config['section']['client_port'] = '9999'

		t = WBroadcastNetworkTransport(
			WNetworkNativeTransportSocketConfig('section', 'client_address', 'client_port'),
			WNetworkNativeTransportSocketConfig('section', 'server_address', 'server_port')
		)

		s = t.create_client_socket(config)
		assert(isinstance(s, socket.socket) is True)
		assert(s.type == socket.SOCK_DGRAM)
		assert(s.getsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST) == 1)


class TestWMulticastNetworkTransport:

	def test_transport(self):
		config = WConfig()
		config.add_section('section')
		config['section']['server_address'] = ''
		config['section']['server_port'] = '7070'
		config['section']['client_address'] = '127.0.0.1'
		config['section']['client_port'] = '7070'

		transport = WMulticastNetworkTransport(
			WNetworkNativeTransportSocketConfig('section', 'client_address', 'client_port'),
			WNetworkNativeTransportSocketConfig('section', 'server_address', 'server_port')
		)

		pytest.raises(ValueError, transport.target_socket, config)
		config['section']['client_address'] = '239.200.1.2'
		assert(isinstance(transport.target_socket(config), WIPV4SocketInfo) is True)

		s1 = transport.server_socket(config)
		assert(isinstance(s1, socket.socket) is True)
		assert(s1.type == socket.SOCK_DGRAM)

		transport.close_server_socket(config)
		s2 = transport.server_socket(config)
		assert(isinstance(s2, socket.socket) is True)
		assert(s1 != s2)



# -*- coding: utf-8 -*-
# wasp_general/datetime.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import time
from datetime import datetime, timezone, timedelta

from wasp_general.verify import verify_type


def local_tz():
	""" Return current system timezone shift from UTC

	:return: datetime.timezone
	"""
	return timezone(timedelta(0, (time.timezone * -1)))


@verify_type(dt=(datetime, None), local_value=bool)
def utc_datetime(dt=None, local_value=True):
	""" Convert local datetime and/or datetime without timezone information to UTC datetime with timezone
	information.

	:param dt: local datetime to convert. If is None, then system datetime value is used
	:param local_value: whether dt is a datetime in system timezone or UTC datetime without timezone information
	:return: datetime in UTC with tz set
	"""
	# TODO: rename local_value to local_tz or in_local_tz
	if dt is None:
		return datetime.now(tz=timezone.utc)

	result = dt
	if result.utcoffset() is None:
		if local_value is False:
			return result.replace(tzinfo=timezone.utc)
		else:
			result = result.replace(tzinfo=local_tz())

	return result.astimezone(timezone.utc)


@verify_type(dt=(datetime, None), utc_value=bool)
def local_datetime(dt=None, utc_value=True):
	""" Convert UTC datetime and/or datetime without timezone information to local datetime with timezone
	information

	:param dt: datetime in UTC to convert. If is None, then system datetime value is used
	:param utc_value: whether dt is a datetime in UTC or in system timezone without timezone information
	:return: datetime for system (local) timezone with tz set
	"""
	# TODO: rename utc_value to utc_tz or in_utc_tz
	if dt is None:
		return datetime.now(tz=local_tz())

	result = dt
	if result.utcoffset() is None:
		if utc_value is False:
			return result.replace(tzinfo=local_tz())
		else:
			result = result.replace(tzinfo=timezone.utc)

	return result.astimezone(local_tz())



# -*- coding: utf-8 -*-
# wasp_general/network/web.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import re

from wasp_general.verify import verify_type, verify_value

from wasp_general.network.web.proto import WWebSessionProto, WWebRequestProto
from wasp_general.network.web.headers import WHTTPHeaders
from wasp_general.network.web.re_statements import http_method_name, http_path, http_version


class WWebRequest(WWebRequestProto):
	""" :class:`.WWebRequestProto` implementation. Class represent HTTP-request descriptor.
	Call :meth:`.WWebRequest.ro` method to create unchangeable copy
	"""

	request_line_re = re.compile(
		'^(' + http_method_name + ') +(' + http_path + ')( +HTTP/(' + http_version + '))?$'
	)
	"""
	Check for HTTP request line. See RFC 2616, Section 5.1


	"""

	@verify_type(session=WWebSessionProto, method=str, path=str, headers=(WHTTPHeaders, None))
	@verify_type(request_data=(bytes, None))
	@verify_value(method=lambda x: len(x) > 0)
	@verify_value(path=lambda x: len(x) > 0)
	def __init__(self, session, method, path, headers=None, request_data=None):
		"""
		Create new request descriptor

		:param session: request origin
		:param method: called HTTP-method
		:param path: called HTTP-path
		"""
		WWebRequestProto.__init__(self)
		self.__session = session
		self.__method = method.upper()
		self.__path = path
		self.__headers = headers
		self.__request_data = request_data
		self.__ro_flag = False

	def session(self):
		""" Return origin session

		:return: WWebSessionProto
		"""
		return self.__session

	def method(self):
		""" Return requested method

		:return: str
		"""
		return self.__method

	def path(self):
		""" Return requested path

		:return: str
		"""
		return self.__path

	def headers(self):
		""" Return request headers

		:return: WHTTPHeaders
		"""
		return self.__headers

	@verify_type(headers=WHTTPHeaders)
	def set_headers(self, headers):
		""" Set headers for request

		:param headers: headers to set
		:return: None
		"""
		if self.__ro_flag:
			raise RuntimeError('Read-only object changing attempt')
		self.__headers = headers

	def request_data(self):
		""" Return request data

		:return: bytes
		"""
		return self.__request_data

	@verify_type(request_data=bytes)
	def set_request_data(self, request_data):
		""" Set payload data for request

		:param request_data: data to set
		:return: None
		"""
		if self.__ro_flag:
			raise RuntimeError('Read-only object changing attempt')
		self.__request_data = request_data

	@classmethod
	@verify_type('paranoid', session=WWebSessionProto)
	@verify_type(request_line=str)
	def parse_request_line(cls, session, request_line):
		""" Parse given request line like 'GET /foo' or 'POST /zzz HTTP/1.0'

		:param session: origin session
		:param request_line: line to parse
		:return: WWebRequest
		"""
		r = cls.request_line_re.search(request_line)
		if r is not None:
			method, path, protocol_sentence, protocol_version = r.groups()
			return WWebRequest(session, method, path)
		raise ValueError('Invalid request line')

	@verify_type('paranoid', http_code=str)
	def parse_headers(self, http_code):
		""" Parse http-code (like 'Header-X: foo\r\nHeader-Y: bar\r\n') and retrieve (save) HTTP-headers

		:param http_code: code to parse
		:return: None
		"""
		if self.__ro_flag:
			raise RuntimeError('Read-only object changing attempt')
		self.__headers = WHTTPHeaders.import_headers(http_code)

	def ro(self):
		""" Create read-only copy

		:return: WWebRequest
		"""
		request = WWebRequest(
			self.session(), self.method(), self.path(),
			headers=self.headers().ro(), request_data=self.request_data()
		)
		request.__ro_flag = True
		return request



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.messenger.proto import WMessengerEnvelopeProto
from wasp_general.network.messenger.envelope import WMessengerEnvelope, WMessengerTextEnvelope, WMessengerBytesEnvelope


class TestWMessengerEnvelope:

	def test(self):
		envelope = WMessengerEnvelope(1)
		assert(isinstance(envelope, WMessengerEnvelope) is True)
		assert(isinstance(envelope, WMessengerEnvelopeProto) is True)
		assert(envelope.message() == 1)
		assert(envelope.meta() == {})

		envelope = WMessengerEnvelope('zxc', meta={'a': 1, 'b': 'c'})
		assert(envelope.message() == 'zxc')
		assert(envelope.meta() == {'a': 1, 'b': 'c'})

		envelope = WMessengerEnvelope(None, meta=envelope)
		assert(envelope.message() is None)
		assert(envelope.meta() == {'a': 1, 'b': 'c'})

		envelope.add_meta('z', 0)
		assert(envelope.meta() == {'a': 1, 'b': 'c', 'z': 0})

		envelope.meta()['q'] = 'q'
		assert(envelope.meta() == {'a': 1, 'b': 'c', 'z': 0})

		class E(WMessengerEnvelopeProto):

			def message(self):
				pass

			def meta(self):
				return {}

		pytest.raises(TypeError, WMessengerEnvelope, 1, E())


class TestWMessengerTextEnvelope:

	def test(self):
		s = WMessengerTextEnvelope('q')
		assert(isinstance(s, WMessengerTextEnvelope) is True)
		assert(isinstance(s, WMessengerEnvelope) is True)

		pytest.raises(TypeError, WMessengerTextEnvelope, None)


class TestWMessengerBytesEnvelope:
	def test(self):
		s = WMessengerBytesEnvelope(b'q')
		assert (isinstance(s, WMessengerBytesEnvelope) is True)
		assert (isinstance(s, WMessengerEnvelope) is True)

		pytest.raises(TypeError, WMessengerBytesEnvelope, None)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.types.binarray import WBinArray


class TestWBinArray:
	def test_init(self):
		pytest.raises(TypeError, WBinArray, 0.1)
		pytest.raises(ValueError, WBinArray, -1)
		pytest.raises(ValueError, WBinArray, 1, 0)
		assert(isinstance(WBinArray(0, 1), WBinArray) is True)

	def test_len(self):
		assert(len(WBinArray(0, 5)) == 5)
		assert(len(WBinArray(0)) == 1)
		assert(len(WBinArray(4)) == 3)

	def test_str(self):
		assert(str(WBinArray(0, 8)) == '00000000')
		assert(str(WBinArray(3)) == '11')
		assert(str(WBinArray(3, 8)) == '00000011')

	def test_int(self):
		assert(int(WBinArray(5)) == 5)

	def test___getitem__(self):
		assert(WBinArray(6)[2] == 0)
		assert(WBinArray(4)[0] == 1)
		pytest.raises(IndexError, WBinArray(3, 8).__getitem__, 10)
		assert(isinstance(WBinArray(3, 8)[5:], WBinArray) is True)
		assert(str(WBinArray(3, 8)[5:]) == '011')
		assert(str(WBinArray(3, 8)[:4]) == '0000')
		assert(str(WBinArray(3, 8)[3:7]) == '0001')
		assert(isinstance(WBinArray(3, 8)[7:], WBinArray) is True)
		pytest.raises(IndexError, WBinArray(3, 8).__getitem__, slice(10, None))
		pytest.raises(IndexError, WBinArray(3, 8).__getitem__, slice(-1, None))

	def test___setitem__(self):
		a = WBinArray(0, 4)
		a[0] = 1
		assert(int(a) == 8)
		a[0] = 0
		assert(int(a) == 0)
		a[0] = 2
		assert(int(a) == 8)
		pytest.raises(TypeError, "a[0] = 0.1")
		pytest.raises(ValueError, "a[0] = -1")
		pytest.raises(IndexError, "a[-1] = 0")
		pytest.raises(IndexError, "a[8] = 0")

		pytest.raises(OverflowError, "a[0] = 20")
		a = WBinArray(0)
		a[0] = 20

	def test_add_sub(self):
		assert(isinstance(WBinArray(0) + 0, int) is True)
		assert(isinstance(WBinArray(0) - 0, int) is True)
		assert((WBinArray(4) + 2) == 6)
		assert((WBinArray(4) + WBinArray(3)) == 7)
		assert((WBinArray(4) - 2) == 2)
		assert((WBinArray(4) - 6) == -2)
		assert((WBinArray(4) - WBinArray(6)) == -2)

	def test_resize(self):
		a = WBinArray(1, 1)
		pytest.raises(IndexError, a.__getitem__, 2)
		a.resize(3)
		assert(a[2] == 1)
		a.resize(1)
		pytest.raises(IndexError, a.__getitem__, 2)
		assert(a[0] == 1)

	def test_split(self):
		assert(int(WBinArray(13).split(2)[0]) == 3)
		assert(int(WBinArray(13).split(2)[1]) == 1)
		assert(int(WBinArray(17).split(2)[0]) == 1)
		assert(int(WBinArray(17).split(2)[1]) == 0)
		assert(int(WBinArray(17).split(2)[2]) == 1)

	def test_concat(self):
		pytest.raises(TypeError, WBinArray(3, 2).concat, 10)
		assert(int(WBinArray(3, 2).concat(WBinArray(1, 2))) == 13)

		pytest.raises(TypeError, WBinArray(3, 2).rconcat, 10)
		assert(int(WBinArray(3, 2).rconcat(WBinArray(1, 2))) == 7)

	def test_extend(self):
		assert(int(WBinArray(3, 2).extend(WBinArray(1, 2), WBinArray(1))) == 27)

	def test_swipe(self):
		assert(int(WBinArray(13, 4).swipe()) == 11)

	def test_join(self):
		assert(WBinArray.join() is None)
		assert(int(WBinArray(3, 2)) == int(WBinArray.join(WBinArray(3, 2))))
		assert(int(WBinArray.join(WBinArray(3, 2), WBinArray(1, 2))) == 13)



# -*- coding: utf-8 -*-
# wasp_general/network/web/cookies.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from http.cookies import SimpleCookie
import re

from wasp_general.verify import verify_type, verify_value
from wasp_general.network.web.re_statements import http_path, http_cookie_expires, http_cookie_max_age
from wasp_general.network.web.re_statements import http_cookie_secure, http_cookie_httponly, http_cookie_domain


class WHTTPCookie:
	""" This class represent a single HTTP Cookie as it is described in RFC 6265.
	Call :meth:`.WHTTPCookie.ro` method to create unchangeable cookie copy
	"""

	cookie_name_non_compliance_re = re.compile(b'.*[\x00-\x1e\x7f()<>@,;:\\\\"/\[\]?={} \t].*')
	"""
	Check for non-acceptable cookie name

	see RFC 6265, Section 4.1.1
	"""

	cookie_value_non_compliance_re = re.compile(b'.*[\x00-\x1e\x7f \",;\\\].*')
	"""
	Check for non-acceptable cookie value

	see RFC 6265, Section 4.1.1
	"""

	cookie_attr_value_compliance = {
		'path': re.compile(http_path),
		'expires': re.compile(http_cookie_expires),
		'max-age': re.compile(http_cookie_max_age),
		'secure': re.compile(http_cookie_secure),
		'httponly': re.compile(http_cookie_httponly),
		'domain': re.compile(http_cookie_domain)
	}
	""" Dictionary of valid attribute names and appropriate values regexp
	"""

	@staticmethod
	def cookie_name_check(cookie_name):
		""" Check cookie name for validity. Return True if name is valid

		:param cookie_name: name to check
		:return: bool
		"""
		cookie_match = WHTTPCookie.cookie_name_non_compliance_re.match(cookie_name.encode('us-ascii'))
		return len(cookie_name) > 0 and cookie_match is None

	@staticmethod
	def cookie_value_check(cookie_value):
		""" Check cookie value for validity. Return True if value is valid

		:param cookie_value: value to check
		:return: bool
		"""
		return WHTTPCookie.cookie_value_non_compliance_re.match(cookie_value.encode('us-ascii')) is None

	@staticmethod
	def cookie_attr_value_check(attr_name, attr_value):
		""" Check cookie attribute value for validity. Return True if value is valid

		:param attr_name: attribute name to check
		:param attr_value: attribute value to check
		:return: bool
		"""
		attr_value.encode('us-ascii')
		return WHTTPCookie.cookie_attr_value_compliance[attr_name].match(attr_value) is not None

	@verify_type(name=str, value=str)
	@verify_value(name=lambda x: WHTTPCookie.cookie_name_check(x))
	@verify_value(value=lambda x: WHTTPCookie.cookie_value_check(x))
	def __init__(self, name, value, **attrs):
		""" Construct new cookie with defined name, value and attributes

		:param name: cookie name
		:param value: cookie value
		:param attrs: cookie attribute with its value. Attribute name must be in lowercase. In order to set
		'max-age' attribute, dash character ('-') may be replaced by underscore ('_') both variants are
		allowed
		"""
		self.__name = name
		self.__value = value
		self.__attrs = {}
		self.__ro_flag = False
		for attr_name in attrs:
			self.attr(attr_name, attrs[attr_name])

	def name(self):
		""" Return cookie name

		:return: str
		"""
		return self.__name

	@verify_type(name=str, new_value=(str, None))
	@verify_value(new_value=lambda x: x is None or WHTTPCookie.cookie_value_check(x))
	def value(self, new_value=None):
		""" Return cookie value. Cookie value can be updated, when new_value is not None. Cookie value
		couldn't be changed if cookie is in read-only mode (RuntimeError exception is raised).

		:param new_value: new value to set
		:return: str
		"""
		if new_value is not None:
			if self.__ro_flag:
				raise RuntimeError('Read-only cookie changing attempt')
			self.__value = new_value
		return self.__value

	def attrs_as_dict(self):
		""" Return cookie attributes as dictionary, where keys are attribute names and values are their
		values

		:return: dict
		"""
		return self.__attrs.copy()

	@verify_type(name=str)
	def __attr_name(self, name):
		""" Return suitable and valid attribute name. This method replaces dash char to underscore. If name
		is invalid ValueError exception is raised

		:param name: cookie attribute name
		:return: str
		"""
		if name not in self.cookie_attr_value_compliance.keys():
			suggested_name = name.replace('_', '-').lower()
			if suggested_name not in self.cookie_attr_value_compliance.keys():
				raise ValueError('Invalid attribute name is specified')
			name = suggested_name
		return name

	@verify_type('paranoid', attr_name=str)
	@verify_type(attr_value=(str, None))
	def attr(self, attr_name, attr_value=None):
		""" Return attribute value. Attribute value can be updated with this method. In order to update
		attribute value attr_value must be set. Cookie attribute value couldn't be changed if cookie is
		in read-only mode (RuntimeError exception is raised).

		:param attr_name: target attribute name
		:param attr_value: new value to set
		:return: str
		"""
		name = self.__attr_name(attr_name)
		if attr_value is not None:
			if WHTTPCookie.cookie_attr_value_check(name, attr_value) is not True:
				raise ValueError('Unacceptable value passed')
			if self.__ro_flag:
				raise RuntimeError('Read-only cookie changing attempt')
			self.__attrs[name] = attr_value
			return attr_value
		return self.__attrs[name]

	@verify_type('paranoid', attr_name=str)
	def remove_attr(self, attr_name):
		""" Remove cookie attribute. Cookie attribute couldn't be removed if cookie is in read-only mode
		(RuntimeError exception is raised).

		:param attr_name: name of attribute to remove
		:return: None
		"""
		if self.__ro_flag:
			raise RuntimeError('Read-only cookie changing attempt')
		name = self.__attr_name(attr_name)
		if name in self.__attrs.keys():
			self.__attrs.pop(attr_name)

	def __str__(self):
		""" Return valid "Set-Cookie" HTTP-header for this cookie

		:return: str
		"""
		simple_cookie = SimpleCookie()
		simple_cookie[self.__name] = self.__value
		for attr_name in self.__attrs.keys():
			simple_cookie[self.__name][attr_name] = self.__attrs[attr_name]
		return str(simple_cookie)

	def ro(self):
		""" Return read-only copy

		:return: WHTTPCookie
		"""
		ro_cookie = self.copy()
		ro_cookie.__ro_flag = True
		return ro_cookie

	def copy(self):
		""" Return copy

		:return: WHTTPCookie
		"""
		copy_cookie = WHTTPCookie(self.__name, self.__value)
		copy_cookie.__attrs = self.__attrs.copy()
		return copy_cookie


class WHTTPCookieJar:
	""" Class represent collection of cookies. Call :meth:`.WHTTPCookieJar.ro` method to create read-only copy
	(in this state no changes are allowed)
	"""

	def __init__(self):
		""" Construct new collection
		"""
		self.__cookies = {}
		self.__ro_flag = False

	def cookies(self):
		""" Return available cookie names

		:return: tuple of str
		"""
		return tuple(self.__cookies.keys())

	@verify_type(cookie=WHTTPCookie)
	def add_cookie(self, cookie):
		""" Add new cookie (or replace if there is cookie with the same name already)

		:param cookie: cookie to add
		:return: None
		"""
		if self.__ro_flag:
			raise RuntimeError('Read-only cookie-jar changing attempt')
		self.__cookies[cookie.name()] = cookie

	@verify_type(cookie_name=str)
	def remove_cookie(self, cookie_name):
		""" Remove cookie by its name

		:param cookie_name: cookie name
		:return:
		"""
		if self.__ro_flag:
			raise RuntimeError('Read-only cookie-jar changing attempt')
		if cookie_name in self.__cookies.keys():
			self.__cookies.pop(cookie_name)

	@verify_type(item=str)
	def __getitem__(self, item):
		""" Get cookie by its name

		:param item: cookie name
		:return:
		"""
		return self.__cookies[item]

	def __iter__(self):
		""" Iterate over cookie collection

		:return: None
		"""
		for cookie in self.__cookies.values():
			yield cookie

	def ro(self):
		""" Return read-only copy

		:return: WHTTPCookieJar
		"""
		ro_jar = WHTTPCookieJar()
		for cookie in self.__cookies.values():
			ro_jar.add_cookie(cookie.ro())
		ro_jar.__ro_flag = True
		return ro_jar

	@classmethod
	@verify_type(simple_cookie=SimpleCookie)
	def import_simple_cookie(cls, simple_cookie):
		""" Create cookie jar from SimpleCookie object

		:param simple_cookie: cookies to import
		:return: WHTTPCookieJar
		"""
		cookie_jar = WHTTPCookieJar()
		for cookie_name in simple_cookie.keys():
			cookie_attrs = {}
			for attr_name in WHTTPCookie.cookie_attr_value_compliance.keys():
				attr_value = simple_cookie[cookie_name][attr_name]
				if attr_value != '':
					cookie_attrs[attr_name] = attr_value

			cookie_jar.add_cookie(WHTTPCookie(
				cookie_name, simple_cookie[cookie_name].value, **cookie_attrs
			))
		return cookie_jar

	@classmethod
	@verify_type(cookie_text=str)
	def import_header_text(cls, cookie_text):
		""" Create cookie jar from HTTP Header text like 'Set-Cookie: cookie=value'

		:param cookie_text: http header code
		:return: WHTTPCookieJar
		"""
		simple_cookie = SimpleCookie()
		simple_cookie.load(cookie_text)
		return cls.import_simple_cookie(simple_cookie)



# -*- coding: utf-8 -*-
# wasp_general/verfy.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import sys
import os
from inspect import getfullargspec, isclass, isfunction, getsource
from decorator import decorator


class Verifier:
	""" Base class for verifier implementation.

	Verifiers are classes, that generates decorators (which are later used for runtime function arguments
	checking). Derived classes (such as :class:`.TypeVerifier`, :class:`SubclassVerifier` and
	:class:`ValueVerifier`) check arguments for type and/or value validity. But each derived class uses
	its own syntax for check declaration (see :meth:`.Verifier.check`).

	Same checks can be grouped into one sentence if they are used for different arguments. Each statement can
	be marked by tag or tags for runtime disabling. If statement doesn't have tag then it always run checks.

	Checks can be simple (implemented by lambda-statements) or complex
	(implemented by standalone functions or classes). Because target function is decorated for checking it is
	possible to run checks sequentially.

	Example: ::

		@verify_type(a=int, b=str, d=(int, None), e=float)
		@verify_subclass(c=A, args=B)
		@verify_value(a=(lambda x: x > 5, lambda x: x < 10), args=lambda x: x > 0)
		@verify_value(c=lambda x: x.a == 'foo', d=lambda x: x is None or x < 0)
		def foo(a, b, c, d=None, *args, **kwargs):
			pass
	"""

	__default_environment_var__ = 'WASP_VERIFIER_DISABLE_CHECKS'
	""" Default environment variable name that is used for check bypassing. Variable must contain tags separated by \
	:attr:`.Verifier.__tags_delimiter__`. To bypass certain check all of its tags must be defined in variable.

	Currently used values:
		'paranoid' - for excess checks, like that test parameters, that doesn't processed by the specific
		function or is passing them to other function that has checks
	"""

	__tags_delimiter__ = ':'
	""" String that is used for tag separation :attr:`.Verifier.__default_environment_var__`"""

	def __init__(self, *tags, env_var=None, silent_checks=False):
		"""Construct a new :class:`.Verifier`

		:param tags: Tags to mark this checks. Now only strings are suitable.
		:param env_var: Environment variable name that is used for check bypassing. If is None, then default  \
		value is used :attr:`.Verifier.__default_environment_var__`
		:param silent_checks: If it is not True, then debug information will be printed to stderr.
		"""
		self._tags = list(tags)
		self._env_var = env_var if env_var is not None else self.__class__.__default_environment_var__
		self._silent_checks = silent_checks

	def decorate_disabled(self):
		""" Return True if this decoration must be omitted, otherwise - False.
		This class searches for tags values in environment variable
		(:attr:`.Verifier.__default_environment_var__`), Derived class can implement any logic

		:return: bool
		"""
		if self._env_var not in os.environ or len(self._tags) == 0:
			return False
		env = os.environ[self._env_var].split(self.__class__.__tags_delimiter__)
		for tag in self._tags:
			if tag not in env:
				return False
		return True

	def check(self, arg_spec, arg_name, decorated_function):
		"""Return callable object that takes single value - future argument. This callable must raise
		an exception if error occurs. It is recommended to return None if everything is OK

		:param arg_spec: specification that is used to describe check like types, lambda-functions, \
		list of types just anything (see :meth:`.Verifier.decorator`)
		:param arg_name: parameter name from decorated function
		:param decorated_function: target function (function to decorate)

		:return: None
		"""
		return lambda x: None

	def _args_checks_gen(self, decorated_function, function_spec, arg_specs):
		""" Generate checks for positional argument testing

		:param decorated_function: function decorator
		:param function_spec: function inspect information
		:param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)

		:return: internal structure, that is used by :meth:`.Verifier._args_checks_test`
		"""
		inspected_args = function_spec.args
		args_check = {}

		for i in range(len(inspected_args)):
			arg_name = inspected_args[i]

			if arg_name in arg_specs.keys():
				args_check[arg_name] = self.check(arg_specs[arg_name], arg_name, decorated_function)
		return args_check

	def _args_checks_test(self, original_function, function_spec, checks, args, arg_specs):
		""" Test positional arguments by a generated checks

		:param original_function: original function, that was decorated
		:param function_spec: function inspect information
		:param checks: tests, that was generated by the :meth:`.Verifier._args_checks_gen`
		:param args: positional arguments to check (with varargs)
		:param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)

		:return: None
		"""
		inspected_args = function_spec.args

		for i in range(len(inspected_args)):
			param_name = inspected_args[i]
			if param_name in checks:
				try:
					checks[param_name](args[i])
				except Exception as e:
					self.help_info(e, original_function, param_name, arg_specs[param_name])
					raise

	def _varargs_checks_gen(self, decorated_function, function_spec, arg_specs):
		""" Generate checks for positional variable argument (varargs) testing

		:param decorated_function: function decorator
		:param function_spec: function inspect information
		:param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)

		:return: internal structure, that is used by :meth:`.Verifier._varargs_checks_test`
		"""
		inspected_varargs = function_spec.varargs

		if inspected_varargs is not None and inspected_varargs in arg_specs.keys():
			return self.check(
				arg_specs[inspected_varargs], inspected_varargs, decorated_function
			)

	def _varargs_checks_test(self, original_function, function_spec, check, args, arg_specs):
		""" Test varargs by a generated check

		:param original_function: original function, that was decorated
		:param function_spec: function inspect information
		:param check: test, that was generated by the :meth:`.Verifier._varargs_checks_gen`
		:param args: positional arguments to check (all of them)
		:param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)

		:return: None
		"""
		inspected_args = function_spec.args
		inspected_varargs = function_spec.varargs

		if inspected_varargs is not None and check is not None:
			for i in range(len(inspected_args), len(args)):
				try:
					check(args[i])
				except Exception as e:
					self.help_info(e, original_function, inspected_varargs, arg_specs[inspected_varargs])
					raise

	def _kwargs_checks_gen(self, decorated_function, function_spec, arg_specs):
		""" Generate checks for keyword argument testing

		:param decorated_function: function decorator
		:param function_spec: function inspect information
		:param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)

		:return: internal structure, that is used by :meth:`.Verifier._kwargs_checks_test`
		"""
		args_names = []
		args_names.extend(function_spec.args)
		if function_spec.varargs is not None:
			args_names.append(function_spec.args)

		args_check = {}

		for arg_name in arg_specs.keys():
			if arg_name not in args_names:
				args_check[arg_name] = self.check(
					arg_specs[arg_name], arg_name, decorated_function
				)
		return args_check

	def _kwargs_checks_test(self, original_function, checks, kwargs, arg_specs):
		""" Test keyword arguments by a generated checks

		:param original_function: original function, that was decorated
		:param checks: tests, that was generated by the :meth:`.Verifier._kwargs_checks_gen`
		:param kwargs: keyword arguments to check
		:param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)

		:return: None
		"""

		for kw_key, kw_value in kwargs.items():
			if kw_key in checks.keys():
				try:
					checks[kw_key](kw_value)
				except Exception as e:
					self.help_info(e, original_function, kw_key, arg_specs[kw_key])
					raise

	def decorator(self, **arg_specs):
		""" Return decorator that can decorate target function

		:param arg_specs: dictionary where keys are parameters name and values are theirs specification.\
		Specific specification is passed as is to :meth:`Verifier.check` method with corresponding \
		parameter name.

		:return: function
		"""

		if self.decorate_disabled() is True:
			def empty_decorator(decorated_function):
				return decorated_function
			return empty_decorator

		def first_level_decorator(decorated_function):

			function_spec = getfullargspec(decorated_function)
			args_checks = self._args_checks_gen(decorated_function, function_spec, arg_specs)
			varargs_check = self._varargs_checks_gen(decorated_function, function_spec, arg_specs)
			kwargs_checks = self._kwargs_checks_gen(decorated_function, function_spec, arg_specs)

			def second_level_decorator(original_function, *args, **kwargs):
				self._args_checks_test(original_function, function_spec, args_checks, args, arg_specs)
				self._varargs_checks_test(original_function, function_spec, varargs_check, args, arg_specs)
				self._kwargs_checks_test(original_function, kwargs_checks, kwargs, arg_specs)

				return original_function(*args, **kwargs)
			return decorator(second_level_decorator)(decorated_function)
		return first_level_decorator

	def help_info(self, exc, decorated_function, arg_name, arg_spec):
		""" Print debug information to stderr. (Do nothing if object was constructed with silent_checks=True)

		:param exc: raised exception
		:param decorated_function: target function (function to decorate)
		:param arg_name: function parameter name
		:param arg_spec: function parameter specification

		:return: None
		"""
		if self._silent_checks is not True:
			print('Exception raised:', file=sys.stderr)
			print(str(exc), file=sys.stderr)
			fn_name = Verifier.function_name(decorated_function)
			print('Decorated function: %s' % fn_name, file=sys.stderr)
			if decorated_function.__doc__ is not None:
				print('Decorated function docstrings:', file=sys.stderr)
				print(decorated_function.__doc__, file=sys.stderr)
			print('Argument "%s" specification:' % arg_name, file=sys.stderr)
			if isfunction(arg_spec):
				print(getsource(arg_spec), file=sys.stderr)
			else:
				print(str(arg_spec), file=sys.stderr)
			print('', file=sys.stderr)

	@staticmethod
	def function_name(fn):
		""" Return function name in pretty style

		:param fn: source function
		:return: str
		"""
		fn_name = fn.__name__
		if hasattr(fn, '__qualname__'):
			return fn.__qualname__
		elif hasattr(fn, '__self__'):
			owner = fn.__self__
			if isclass(owner) is False:
				owner = owner.__class__
			return '%s.%s' % (owner.__name__, fn_name)
		return fn_name


class TypeVerifier(Verifier):
	""" Verifier that is used for type verification. Checks parameter if it is instance of specified class or
	classes. Specification accepts type or list/tuple/set of types

	Example: ::

		@verify_type(a=int, b=str, d=(int, None), e=float)
		def foo(a, b, c, d=None, **kwargs):
			pass
	"""

	def check(self, type_spec, arg_name, decorated_function):
		""" Return callable that checks function parameter for type validity. Checks parameter if it is
		instance of specified class or classes

		:param type_spec: type or list/tuple/set of types
		:param arg_name: function parameter name
		:param decorated_function: target function
		:return: function
		"""

		def raise_exception(x_spec):
			exc_text = 'Argument "%s" for function "%s" has invalid type' % (
				arg_name, Verifier.function_name(decorated_function)
			)
			exc_text += ' (%s should be %s)' % (x_spec, type_spec)
			raise TypeError(exc_text)

		if isinstance(type_spec, (tuple, list, set)):
			for single_type in type_spec:
				if (single_type is not None) and isclass(single_type) is False:
					raise RuntimeError(
						'Invalid specification. Must be type or tuple/list/set of types'
					)
			if None in type_spec:
				type_spec = tuple(filter(lambda x: x is not None, type_spec))
				return lambda x: None if x is None or isinstance(x, tuple(type_spec)) is True else \
					raise_exception(str((type(x))))
			else:
				return lambda x: None if isinstance(x, tuple(type_spec)) is True else \
					raise_exception(str((type(x))))
		elif isclass(type_spec):
			return lambda x: None if isinstance(x, type_spec) is True else \
				raise_exception(str((type(x))))
		else:
			raise RuntimeError('Invalid specification. Must be type or tuple/list/set of types')


class SubclassVerifier(Verifier):
	""" Verifier that is used for type verification. Checks parameter if it is class or subclass of
	specified class or classes. Specification accepts type or list/tuple/set of types

	Example: ::

		@verify_subclass(c=A, e=(A, D))
		def foo(a, b, c, d=None, **kwargs):
			pass
	"""

	def check(self, type_spec, arg_name, decorated_function):
		""" Return callable that checks function parameter for class validity. Checks parameter if it is
		class or subclass of specified class or classes

		:param type_spec: type or list/tuple/set of types
		:param arg_name: function parameter name
		:param decorated_function: target function
		:return: function
		"""

		def raise_exception(text_spec):
			exc_text = 'Argument "%s" for function "%s" has invalid type' % (
				arg_name, Verifier.function_name(decorated_function)
			)
			exc_text += ' (%s)' % text_spec
			raise TypeError(exc_text)

		if isinstance(type_spec, (tuple, list, set)):
			for single_type in type_spec:
				if (single_type is not None) and isclass(single_type) is False:
					raise RuntimeError(
						'Invalid specification. Must be type or tuple/list/set of types'
					)
			if None in type_spec:
				type_spec = tuple(filter(lambda x: x is not None, type_spec))
				return lambda x: None if \
					x is None or (isclass(x) is True and issubclass(x, type_spec) is True) else \
					raise_exception(str(x))
			else:
				return lambda x: None if (isclass(x) is True and issubclass(x, type_spec) is True) else \
					raise_exception(str(x))
		elif isclass(type_spec):
			return lambda x: None if (isclass(x) is True and issubclass(x, type_spec) is True) else \
				raise_exception(str(x))
		else:
			raise RuntimeError('Invalid specification. Must be type or tuple/list/set of types')


class ValueVerifier(Verifier):
	""" Verifier that is used for value verification. Checks parameter if its value passes specified restrictions.
	Specification accepts function or list/tuple/set of functions. Each function must accept one parameter and
	must return True or False if it passed restrictions or not.

	Example: ::

		@verify_value(a=(lambda x: x > 5, lambda x: x < 10))
		@verify_value(c=lambda x: x.a == 'foo', d=lambda x: x is None or x < 0)
		def foo(a, b, c, d=None, **kwargs):
		pass
	"""

	def check(self, value_spec, arg_name, decorated_function):
		""" Return callable that checks function parameter for value validity. Checks parameter if its value
		passes specified restrictions.

		:param value_spec: function or list/tuple/set of functions. Each function must accept one parameter and \
		must return True or False if it passed restrictions or not.
		:param arg_name: function parameter name
		:param decorated_function: target function
		:return: function
		"""

		def raise_exception(text_spec):
			exc_text = 'Argument "%s" for function "%s" has invalid value' % (
				arg_name, Verifier.function_name(decorated_function)
			)
			exc_text += ' (%s)' % text_spec
			raise ValueError(exc_text)

		if isinstance(value_spec, (tuple, list, set)):

			for single_value in value_spec:
				if isfunction(single_value) is False:
					raise RuntimeError(
						'Invalid specification. Must be function or tuple/list/set of functions'
					)

			def check(x):
				for f in value_spec:
					if f(x) is not True:
						raise_exception(str(x))

			return check

		elif isfunction(value_spec):
			return lambda x: None if value_spec(x) is True else raise_exception(str(x))
		else:
			raise RuntimeError('Invalid specification. Must be function or tuple/list/set of functions')


def verify_type(*tags, **type_kwargs):
	""" Shortcut for :class:`.TypeVerifier`

	:param tags: verification tags. See :meth:`.Verifier.__init__`
	:param type_kwargs: verifier specification. See :meth:`.TypeVerifier.check`
	:return: decorator (function)
	"""
	return TypeVerifier(*tags).decorator(**type_kwargs)


def verify_subclass(*tags, **type_kwargs):
	""" Shortcut for :class:`.SubclassVerifier`

	:param tags: verification tags. See :meth:`.Verifier.__init__`
	:param type_kwargs: verifier specification. See :meth:`.SubclassVerifier.check`
	:return: decorator (function)
	"""
	return SubclassVerifier(*tags).decorator(**type_kwargs)


def verify_value(*tags, **type_kwargs):
	""" Shortcut for :class:`.ValueVerifier`

	:param tags: verification tags. See :meth:`.Verifier.__init__`
	:param type_kwargs: verifier specification. See :meth:`.ValueVerifier.check`
	:return: decorator (function)
	"""
	return ValueVerifier(*tags).decorator(**type_kwargs)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.crypto.rsa import WRSA


class TestWRSA:

	def test_generate_private(self):
		assert(WRSA.generate_private() is not None)
		assert(WRSA.generate_private(2048).size() == 2047)
		assert(WRSA.generate_private(1280).size() == 1279)
		assert(WRSA.generate_private(1024).size() == 1023)

	def test_export_key(self):
		pk = WRSA.generate_private(1024)
		assert(isinstance(WRSA.export_key(pk), bytes) is True)
		assert(WRSA.export_key(pk) == WRSA.export_key(pk))
		assert(WRSA.export_key(pk, 'foo') != WRSA.export_key(pk, 'foo'))
		assert(WRSA.export_key(pk, b'foo') != WRSA.export_key(pk, b'foo'))
		pytest.raises(TypeError, WRSA.export_key, pk, 1)

		pytest.raises(TypeError, WRSA.import_key, 1)
		pytest.raises(TypeError, WRSA.import_key, WRSA.export_key(pk), 1)
		assert(WRSA.import_key(WRSA.export_key(pk)) == pk)
		assert(WRSA.import_key(WRSA.export_key(pk, 'foo'), 'foo') == pk)
		pytest.raises(ValueError, WRSA.import_key, WRSA.export_key(pk, 'foo'))
		assert(WRSA.import_key(WRSA.export_key(pk), 'foo') == pk)

		pk = WRSA.generate_public(pk)
		assert(WRSA.export_key(pk) == WRSA.export_key(pk))
		assert(WRSA.export_key(pk, 'foo') == WRSA.export_key(pk, 'foo'))
		assert(WRSA.export_key(pk, b'foo') == WRSA.export_key(pk, b'foo'))
		assert(WRSA.export_key(pk, b'foo') == WRSA.export_key(pk, 'bar'))
		assert(WRSA.import_key(WRSA.export_key(pk)) == pk)

	def test_generate_public(self):
		pytest.raises(TypeError, WRSA.generate_public, 'bla-bla')

		assert(WRSA.generate_public(WRSA.generate_private(1024)).size() == 1023)
		assert(WRSA.generate_public(WRSA.generate_private(2048)).size() == 2047)
		assert(WRSA.generate_public(WRSA.generate_private(1280)).size() == 1279)
		pk = WRSA.generate_private(1024)
		pk = WRSA.generate_public(pk)
		assert(pk is not None)
		assert(pk == WRSA.generate_public(pk))
		assert(WRSA.export_key(pk) == WRSA.export_key(WRSA.generate_public(pk)))

	def test_encrypt(self):
		pk = WRSA.generate_private(1024)
		pytest.raises(TypeError, WRSA.encrypt)
		pytest.raises(TypeError, WRSA.encrypt, b'')
		pytest.raises(TypeError, WRSA.encrypt, '', pk)

		pytest.raises(TypeError, WRSA.decrypt)
		pytest.raises(TypeError, WRSA.decrypt, b'')
		pytest.raises(TypeError, WRSA.decrypt, '', pk)

		assert(WRSA.encrypt('qwerty'.encode(), pk) != WRSA.encrypt('qwerty'.encode(), pk))
		assert(WRSA.encrypt('qwerty'.encode(), pk) != WRSA.encrypt('qwerty'.encode(), WRSA.generate_public(pk)))
		assert(WRSA.decrypt(WRSA.encrypt('qwerty'.encode(), pk), pk) == b'qwerty')
		assert(WRSA.decrypt(WRSA.encrypt('qwerty'.encode(), WRSA.generate_public(pk)), pk) == b'qwerty')

	def test_string_encrypt(self):
		pk = WRSA.generate_private(1024)
		pytest.raises(TypeError, WRSA.string_encrypt)
		pytest.raises(TypeError, WRSA.string_encrypt, '')
		pytest.raises(TypeError, WRSA.string_encrypt, b'')
		pytest.raises(TypeError, WRSA.string_encrypt, b'', pk)

		assert(WRSA.string_encrypt('qwerty', WRSA.generate_public(pk)) != WRSA.encrypt('qwerty'.encode(), pk))

		pytest.raises(TypeError, WRSA.string_decrypt)
		pytest.raises(TypeError, WRSA.string_decrypt, b'')
		pytest.raises(TypeError, WRSA.string_decrypt, '')
		pytest.raises(TypeError, WRSA.string_decrypt, '', pk)
		assert(WRSA.string_decrypt(WRSA.string_encrypt('qwerty', pk), pk) == 'qwerty')

		koi_string = WRSA.encrypt('халё'.encode('koi8-r'), pk)
		assert(WRSA.string_decrypt(koi_string, pk, text_encoding='koi8-r') == 'халё')
		pytest.raises(TypeError, WRSA.string_decrypt, koi_string, pk, text_encoding=1)



# -*- coding: utf-8 -*-
# wasp_general/network/web/service.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import re
from inspect import getfullargspec, ismethod
from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type, verify_subclass, verify_value
from wasp_general.network.web.proto import WWebTargetRouteProto, WWebRouteMapProto, WWebRequestProto, WWebServiceProto
from wasp_general.network.web.proto import WWebPresenter, WWebErrorPresenter, WWebSessionProto
from wasp_general.network.web.proto import WWebPresenterCollectionProto, WWebPresenterFactoryProto
from wasp_general.network.web.re_statements import wasp_route_arg_name_selection, wasp_route_arg_value_selection
from wasp_general.network.web.re_statements import wasp_route_custom_arg_selection, wasp_route_import_pattern
from wasp_general.network.web.re_statements import wasp_route_arg_name
from wasp_general.network.web.response import WWebResponse
from wasp_general.network.web.headers import WHTTPHeaders
from wasp_general.network.web.debug import WWebDebugInfo


class WSimpleErrorPresenter(WWebErrorPresenter):
	""" Simple :class:`.WWebErrorPresenter` implementation
	"""

	@verify_type('paranoid', request=WWebRequestProto)
	def __init__(self, request):
		""" Create new error presenter

		:param request: origin request
		"""
		WWebErrorPresenter.__init__(self, request)
		self.__messages = {
			404: 'Page not found'
		}

	@verify_type(code=int)
	@verify_value(code=lambda x: x > 0)
	def __message__(self, code):
		if code in self.__messages.keys():
			return self.__messages[code]
		return 'Internal error: %i' % code

	@verify_type('paranoid', code=int)
	@verify_value('paranoid', code=lambda x: x > 0)
	def error_code(self, code):
		""" :meth:`.WWebErrorPresenter.error_code` method implementation
		"""
		return WWebResponse(
			status=code,
			headers=WHTTPHeaders(**{'Content-Type': 'text/plain; charset=utf-8'}),
			response_data=self.__message__(code).encode()
		)

	@verify_type(exception=Exception)
	def exception_error(self, exception):
		""" :meth:`.WWebErrorPresenter.exception_error` method implementation
		"""
		return self.error_code(500)

	@classmethod
	def __presenter_name__(cls):
		""" :meth:`.WWebPresenter.__presenter_name__` method implementation
		"""
		return 'simple-error-presenter'


class WWebRoute:
	""" Class is used for single route description. It helps to match single request to corresponding presenter
	(:class:`.WWebPresenter`). Every route is described by a single pattern (that later compiles to regular
	expression). Pattern may define different route arguments. There is a way to define default arguments with
	default values, this is done by passing arguments in object constructor. Every route argument
	(except special arguments) will be provided to the related presenter action via :class:`.WWebTargetRoute` class.

	Every special argument can be defined in a object constructor, but there is a single special argument
	('action') that can be defined in pattern also. Special arguments usage:
	- 'action': redefines presenter action. Default value is 'index'
	- 'virtual_hosts': is used for request matching. If defined, there must be 'Host' header
	in client request, which equals hostname saved in 'virtual_hosts'. Hostnames are case-insensitive. Default value
	is None which is matched to any hostname
	- 'protocols': defines protocol for client request matching. Suitable protocols are 'http' or 'https'. Default
	value contains both protocols.
	- 'ports': list/tuple/set of ports client is connected to. Default value is None, which is matched to any port
	- 'methods': available client methods like GET, POST, HEAD. Method name is case-sensitive. Default value allow
	to use GET and POST methods
	"""

	class ArgSearch(metaclass=ABCMeta):
		""" Class for internal usage only. It is used for route argument searching from the given pattern and
		later pattern preparation
		"""

		@abstractmethod
		def args(self):
			""" Return arguments, that were found inside the given pattern. Each argument is a tuple of
			argument name and argument position index inside the origin pattern.

			:return: tuple of (str, int)
			"""
			raise NotImplementedError('This method is abstract')

		@abstractmethod
		def reduce_pattern(self):
			""" Return functor that accept single parameter, which is a pattern to prepare. Pattern
			preparation is a process which replace argument specification with regular expression substring,
			regular expression that is used for determine argument value.

			:return: function or lambda
			"""
			raise NotImplementedError('This method is abstract')

	class BasicArgSearch(ArgSearch):
		""" This is a simple route argument search, that is looking for simple pattern "{<argument_name>}".
		In the prepared patter this argument specification will be replaced with "(\w+)" statement
		"""

		arg_re = re.compile(wasp_route_arg_name_selection)
		""" Argument name pattern
		"""

		@verify_type(pattern=str)
		@verify_value(pattern=lambda x: len(x) > 0)
		def __init__(self, pattern):
			""" Construct new search

			:param pattern: pattern with argument specification
			"""
			self.__pattern = pattern
			self.__replace_with = wasp_route_arg_value_selection

		def args(self):
			""" :meth:`.WWebRoute.ArgSearch.args` implementation
			"""
			result = []
			for arg in WWebRoute.BasicArgSearch.arg_re.findall(self.__pattern):
				result.append((arg, self.__pattern.find('{%s}' % arg)))
			return tuple(result)

		def reduce_pattern(self):
			""" :meth:`.WWebRoute.ArgSearch.reduce_pattern` implementation
			"""
			return lambda x: WWebRoute.BasicArgSearch.arg_re.sub(self.__replace_with, x)

	class CustomArgSearch(ArgSearch):
		""" This is more complex argument search than :class:`.WWebRoute.BasicArgSearch`. This class provides
		an ability to control target argument regular expression. Instead of using simple regexp like
		:class:`.WWebRoute.BasicArgSearch` does (:class:`.WWebRoute.BasicArgSearch` uses '(\w+)'), this class
		will search for a special statements that describe and argument name and corresponding regexp.
		Statements obey the following format "{<argument_name>:"<regular_expression>"}".
		"""

		custom_arg_re = re.compile(wasp_route_custom_arg_selection)
		""" Argument name pattern
		"""

		@verify_type(arg_name=str, arg_pos=int, arg_spec=str, arg_custom_re_text=str)
		@verify_value(arg_name=lambda x: len(x) > 0, arg_pos=lambda x: x >= 0)
		@verify_value(arg_spec=lambda x: len(x) > 0, arg_custom_re_text=lambda x: len(x) > 0)
		def __init__(self, arg_name, arg_pos, arg_spec, arg_custom_re_text):
			""" Construct new search. Actually this is a result of search. all the magic happens in
			:meth:`.WWebRoute.CustomArgSearch.custom` method

			:param arg_name: found argument name
			:param arg_pos: found argument position
			:param arg_spec: found argument specification (whole sentence)
			:param arg_custom_re_text: found regular expression
			"""
			self.__arg_name = arg_name
			self.__arg_pos = arg_pos
			self.__arg_spec = arg_spec
			self.__arg_re = arg_custom_re_text

		def args(self):
			""" :meth:`.WWebRoute.ArgSearch.args` implementation
			"""
			return (self.__arg_name, self.__arg_pos),

		def reduce_pattern(self):
			""" :meth:`.WWebRoute.ArgSearch.reduce_pattern` implementation
			"""
			return lambda x: x.replace(self.__arg_spec, self.__arg_re)

		@classmethod
		@verify_type(pattern=str)
		@verify_value(pattern=lambda x: len(x) > 0)
		def custom(cls, pattern):
			""" Parse the given pattern for arguments and iterate over results
			(:class:`.WWebRoute.CustomArgSearch` is yielded)

			:param pattern: pattern for search
			:return: None
			"""
			reduced_pattern = pattern
			for arg in WWebRoute.CustomArgSearch.custom_arg_re.findall(pattern):
				arg_search = WWebRoute.CustomArgSearch(arg[1], pattern.find(arg[0]), arg[0], arg[2])
				reduced_pattern = arg_search.reduce_pattern()(reduced_pattern)
				yield arg_search

	multiple_slashes_re = re.compile("//+")
	""" Regexp for removing extra slashes
	"""

	pattern_parentheses_removing_re = re.compile('(\([^()]*\([^()]*\)[^()]*\))')
	""" Regexp for removing internal parentheses
	"""

	pattern_parentheses_removing_nested_re = re.compile('(\([^()]*\([^()]*\)[^()]*\([^()]*\)[^()]*\))')
	""" Regexp for removing internal nested parentheses
	"""

	pattern_recombination_re = re.compile('(\([^()]*\))')
	""" Regexp for source argument compilation
	"""

	@verify_type(pattern=str, presenter=str, action=(str, None), virtual_hosts=(list, tuple, set, None))
	@verify_type(protocols=(list, tuple, set, None), ports=(list, tuple, set, None))
	@verify_type(methods=(list, tuple, set, None))
	def __init__(self, pattern, presenter, **kwargs):
		""" Create new route

		:param pattern: route declarative pattern, that may contain arguments definition
		:param presenter: target presenter name
		:param kwargs: route arguments (default and special arguments)
		"""

		self.original_pattern = pattern
		## Source pattern. It doesn't change during lifetime, the only usage is debug

		self.route_args = []
		## Route arguments encoded in url

		self.presenter = presenter
		## Presenter to use

		self.presenter_args = {}
		## Default arguments for the given action

		self.action = "index"
		## Action to run (default "index")

		self.virtual_hosts = None
		## List of hosts in URI (None - any host)

		self.ports = None
		## List of ports in URI (None - any port)

		self.protocols = ("http", "https")
		## List of protocols (default "http" and "https")

		self.methods = ("GET", "POST")
		## List of HTTP method (default "GET" and "POST")

		arg_search = WWebRoute.BasicArgSearch(self.normalize_uri(pattern))
		route_args = list(arg_search.args())
		reduce_fn = [arg_search.reduce_pattern()]
		# process simple arguments like {action} or {index}

		for arg_search in WWebRoute.CustomArgSearch.custom(pattern):
			route_args.extend(arg_search.args())
			reduce_fn.append(arg_search.reduce_pattern())
		# process customized arguments like {param1: "(\d+)"}

		route_args.sort(key=lambda x: x[1])
		self.route_args = [x[0] for x in route_args]
		# reorder route args

		pattern = self.normalize_uri(pattern)
		for fn in reduce_fn:
			pattern = fn(pattern)
		self.pattern = "^" + pattern + "$"
		self.re_pattern = re.compile(self.pattern)
		# pattern compilation

		for key in kwargs.keys():

			if key == "action":
				self.action = kwargs["action"]
				continue

			if key == "virtual_hosts":
				self.virtual_hosts = tuple([x.lower() for x in kwargs["virtual_hosts"]])
				continue

			if key == "protocols":
				self.protocols = tuple(kwargs["protocols"])
				continue

			if key == "ports":
				self.ports = tuple(kwargs["ports"])
				continue

			if key == "methods":
				self.methods = tuple(kwargs["methods"])
				continue

			self.presenter_args[key] = kwargs[key]
		# extra presenter args

	@classmethod
	@verify_type(uri=str)
	def normalize_uri(cls, uri):
		""" Normalize the given URI (removes extra slashes)

		:param uri: uri to normalize
		:return: str
		"""
		uri = WWebRoute.multiple_slashes_re.sub("/", uri)

		# remove last slash
		if len(uri) > 1:
			if uri[-1] == '/':
				uri = uri[:-1]

		return uri

	@verify_type(request=WWebRequestProto, service=WWebServiceProto)
	def match(self, request, service):
		""" Check this route for matching the given request. If this route is matched, then target route is
		returned.

		:param request: request to match
		:param service: source service
		:return: WWebTargetRoute or None
		"""
		uri = self.normalize_uri(request.path())

		if request.session().protocol() not in self.protocols:
			return

		if request.method() not in self.methods:
			return

		if self.virtual_hosts and request.virtual_host() not in self.virtual_hosts:
			return

		if self.ports and int(request.session().server_address().port()) not in self.ports:
			return

		match_obj = self.re_pattern.match(uri)

		if not match_obj:
			return

		presenter_action = self.action
		presenter_args = self.presenter_args.copy()

		for i in range(len(self.route_args)):
			if self.route_args[i] == 'action':
				presenter_action = match_obj.group(i + 1)
			else:
				presenter_args[self.route_args[i]] = match_obj.group(i + 1)

		return WWebTargetRoute(self.presenter, presenter_action, self, service.route_map(), **presenter_args)

	def url_for(self, **kwargs):
		""" Generate url for client with specified route arguments. For the source route '/page/{page_index}'
		this method must be called with 'page_index' parameter (for '3' as 'page_index' parameter result will
		be '/page/3' for 'foo_bar' - '/page/foo_bar'). If 'host', 'protocol' and 'ports' are set as single
		element list or they are passed as method parameters, then absolute link is generated as a result.

		Every parameter in the source pattern must be defined. They can be defined as this method parameters
		or in a constructor. If source pattern has more arguments than it can be resolved, then exception will
		be raised

		:param kwargs: route arguments to use
		:return: str
		"""
		host = self.virtual_hosts[0] if self.virtual_hosts is not None and len(self.virtual_hosts) == 1 else None
		port = self.ports[0] if self.ports is not None and len(self.ports) == 1 else None
		protocol = self.protocols[0] if self.protocols is not None and len(self.protocols) == 1 else None
		route_args = self.route_args.copy()

		url_args = []
		for key in kwargs.keys():
			if key == 'host':
				host = kwargs['host']
				continue
			if key == 'port':
				port = kwargs['port']
				continue
			if key == 'protocol':
				protocol = kwargs['protocol']
				continue

			url_args.append(key)

		url_prefix = ''
		if protocol is not None and host is not None:
			url_prefix += ('%s://%s' % (protocol, host))
			if port is not None:
				url_prefix += (':%i' % port)

		pattern = self.pattern[1:-1]
		check = lambda x: WWebRoute.pattern_parentheses_removing_re.search(x) is not None or \
			WWebRoute.pattern_parentheses_removing_nested_re.search(x) is not None
		while check(pattern) is True:
			pattern = WWebRoute.pattern_parentheses_removing_re.sub('()', pattern)
			pattern = WWebRoute.pattern_parentheses_removing_nested_re.sub('(())', pattern)

		arg_i = 0
		while WWebRoute.pattern_recombination_re.search(pattern) is not None:
			arg_name = route_args[arg_i]
			arg_value = None
			if arg_name in url_args:
				arg_value = kwargs[arg_name]
			elif arg_name == 'action':
				arg_value = self.action
			elif arg_name in self.presenter_args:
				arg_value = self.presenter_args[arg_name]

			if arg_value is None:
				raise RuntimeError('Invalid argument')

			pattern = WWebRoute.pattern_recombination_re.sub(arg_value, pattern, 1)
			arg_i += 1

		return url_prefix + self.normalize_uri(pattern)


class WWebTargetRoute(WWebTargetRouteProto):
	""" Simple :class:`.WWebTargetRouteProto` implementation
	"""

	@verify_type(presenter_name=str, presenter_action=str, route=WWebRoute, route_map=WWebRouteMapProto)
	def __init__(self, presenter_name, presenter_action, route, route_map, **presenter_args):
		""" Construct new target route

		:param presenter_name: name of the target presenter. May vary from the presenter name that is specified
		in route object
		:param presenter_action: presenter action (method) to execute
		:param presenter_args: args to execute with
		:param route: source route or None if this object wasn't created within a route
		:param route_map: source route map
		"""
		self.__presenter_name = presenter_name
		self.__presenter_action = presenter_action
		self.__presenter_args = presenter_args
		self.__route = route
		self.__route_map = route_map

	def route(self):
		""" Return origin route map

		:return: WWebRoute
		"""
		return self.__route

	def route_map(self):
		""" Return origin route map

		:return: WWebRouteMap
		"""
		return self.__route_map

	def presenter_name(self):
		""" :meth:`.WWebTargetRouteProto.presenter_class` method implementation
		"""
		return self.__presenter_name

	def presenter_action(self):
		""" :meth:`.WWebTargetRouteProto.presenter_action` method implementation
		"""
		return self.__presenter_action

	def presenter_args(self):
		""" :meth:`.WWebTargetRouteProto.presenter_args` method implementation
		"""
		return self.__presenter_args


class WWebRouteMap(WWebRouteMapProto):
	""" This class represent collection of routes. It has methods for linking a source pattern to th corresponding
	presenter. There is :meth:`WWebRouteMap.route` method that allow to search for a target route by the given
	client request.

	:class:`.WWebRouteMapProto` implementation
	"""

	import_route_re = re.compile(wasp_route_import_pattern)
	""" Regexp for text parsing and custom routes importing
	"""
	import_route_arg_re = re.compile(
		'^\s*(' + wasp_route_arg_name + ')(\s*=\s*(' + wasp_route_arg_value_selection + '))?\s*$'
	)
	""" Regexp for attribute parsing of imported text
	"""

	def __init__(self):
		""" Create new route map
		"""
		self.__routes = []
		self.__error_presenter = WSimpleErrorPresenter

	@verify_type('paranoid', request=WWebRequestProto, service=WWebServiceProto)
	def route(self, request, service):
		""" :meth:`.WWebRouteMapProto.route` method implementation
		"""
		for route in self.__routes:
			result = route.match(request, service)
			if result is not None:
				if self.target_route_valid(result) is True:
					return result

	def error_presenter(self):
		""" :meth:`.WWebRouteMapProto.error_presenter` method implementation
		"""
		return self.__error_presenter

	@verify_subclass(presenter=WWebErrorPresenter)
	def set_error_presenter(self, presenter):
		""" Set error presenter for this route map

		:param presenter: presenter to be used for error handling
		:return: None
		"""
		self.__error_presenter = presenter

	@verify_type('paranoid', pattern=str, presenter=str)
	def connect(self, pattern, presenter, **kwargs):
		""" Connect the given pattern with the given presenter

		:param pattern: URI pattern
		:param presenter: target presenter name
		:param kwargs: route arguments (see :class:`.WWebRoute`)
		:return: None
		"""
		self.__routes.append(WWebRoute(pattern, presenter, **kwargs))

	@verify_type(route=WWebRoute)
	def append(self, route):
		""" Append route to collection

		:param route: route to add
		:return: None
		"""
		self.__routes.append(route)

	@verify_type(route_as_txt=str)
	@verify_value(route_as_txt=lambda x: len(x) > 0)
	def import_route(self, route_as_txt):
		""" Import route written as a string

		:param route_as_txt: single string (single route) to import
		:return: None
		"""
		route_match = WWebRouteMap.import_route_re.match(route_as_txt)
		if route_match is None:
			raise ValueError('Invalid route code')

		pattern = route_match.group(1)
		presenter_name = route_match.group(2)
		route_args = route_match.group(4)  # may be None

		if route_args is not None:
			result_args = {}
			for arg_declaration in route_args.split(","):
				arg_match = WWebRouteMap.import_route_arg_re.match(arg_declaration)
				if arg_match is None:
					raise RuntimeError('Invalid argument declaration in route')
				result_args[arg_match.group(1)] = arg_match.group(3)

			self.connect(pattern, presenter_name, **result_args)
		else:
			self.connect(pattern, presenter_name)


class WSimplePresenterCollection(WWebPresenterCollectionProto):
	""" This is a simple presenter collection

	:class:`.WWebPresenterCollectionProto` implementation
	"""

	def __init__(self):
		""" Construct new collection
		"""
		WWebPresenterCollectionProto.__init__(self)
		self.__presenters = {}

	@verify_subclass(presenter=WWebPresenter)
	def add(self, presenter):
		""" Add presenter to this collection

		:param presenter: presenter to add
		:return: None
		"""
		self.__presenters[presenter.__presenter_name__()] = presenter

	@verify_type(section=str, presenter_name=str)
	@verify_value(section=lambda x: len(x) > 0, presenter_name=lambda x: len(x) > 0)
	def presenter(self, presenter_name):
		""" :meth:`.WWebPresenterCollectionProto.presenter` method implementation
		"""
		if presenter_name in self.__presenters.keys():
			return self.__presenters[presenter_name]

	@verify_type(presenter_name=str)
	@verify_value(presenter_name=lambda x: len(x) > 0)
	def has(self, presenter_name):
		""" :meth:`.WWebPresenterCollectionProto.has` method implementation
		"""
		return presenter_name in self.__presenters.keys()

	def __len__(self):
		""" Return total presenters count

		:return: int
		"""
		return len(self.__presenters)


class WWebService(WWebServiceProto):
	""" This class joins together all web-package functionality.

	:class:`.WWebServiceProto` implementation
	"""

	@verify_type(route_map=(WWebRouteMap, None), collection=(WWebPresenterCollectionProto, None))
	@verify_type(debugger=(WWebDebugInfo, None))
	@verify_subclass(factory=(WWebPresenterFactoryProto, None))
	def __init__(self, route_map=None, collection=None, factory=None, debugger=None):
		""" Create new service

		:param route_map: route map to be uses (if None, internal route map will be used)
		:param collection: collection to be used (if None, internal collection will be used)
		"""
		self.__route_map = route_map if route_map is not None else WWebRouteMap()
		self.__presenter_collection = collection
		self.__factory = factory() if factory is not None else WWebPresenterFactory()
		self.__debugger = debugger
		if self.__presenter_collection is None:
			self.__presenter_collection = WSimplePresenterCollection()

	def route_map(self):
		""" :meth:`.WWebServiceProto.route_map` method implementation
		"""
		return self.__route_map

	def presenter_collection(self):
		""" :meth:`.WWebServiceProto.presenter_collection` method implementation
		"""
		return self.__presenter_collection

	def presenter_factory(self):
		""" :meth:`.WWebServiceProto.presenter_factory` method implementation
		"""
		return self.__factory

	@verify_type(session=WWebSessionProto)
	def process_request(self, session):
		""" Process single request from the given session

		:param session: session for reading requests and writing responses
		:return: None
		"""

		debugger = self.debugger()
		debugger_session_id = debugger.session_id() if debugger is not None else None

		try:
			request = session.read_request()
			if debugger_session_id is not None:
				debugger.request(
					debugger_session_id, request, session.protocol_version(), session.protocol()
				)

			try:
				target_route = self.route_map().route(request, self)
				if debugger_session_id is not None:
					debugger.target_route(debugger_session_id, target_route)

				if target_route is not None:
					response = self.execute(request, target_route)
				else:
					presenter_cls = self.route_map().error_presenter()
					presenter = presenter_cls(request)
					response = presenter.error_code(code=404)

				if debugger_session_id is not None:
					debugger.response(debugger_session_id, response)

			except Exception as e:
				if debugger_session_id is not None:
					debugger.exception(debugger_session_id, e)

				presenter_cls = self.route_map().error_presenter()
				presenter = presenter_cls(request)
				response = presenter.exception_error(e)

			session.write_response(request, response, *response.__pushed_responses__())
		except Exception as e:
			if debugger_session_id is not None:
				debugger.exception(debugger_session_id, e)

			session.session_close()

		if debugger_session_id is not None:
			debugger.finalize(debugger_session_id)

	@verify_type('paranoid', request=WWebRequestProto, target_route=WWebTargetRouteProto)
	def create_presenter(self, request, target_route):
		""" Create presenter from the given requests and target routes

		:param request: client request
		:param target_route: route to use
		:return: WWebPresenter
		"""
		presenter_name = target_route.presenter_name()
		if self.presenter_collection().has(presenter_name) is False:
			raise RuntimeError('No such presenter: %s' % presenter_name)
		presenter_class = self.presenter_collection().presenter(presenter_name)
		return self.presenter_factory().instantiate(presenter_class, request, target_route, self)

	@verify_type('paranoid', request=WWebRequestProto, target_route=WWebTargetRouteProto)
	def execute(self, request, target_route):
		""" :meth:`.WWebServiceProto.execute` method implementation
		"""
		presenter = self.create_presenter(request, target_route)
		presenter_name = target_route.presenter_name()
		action_name = target_route.presenter_action()
		presenter_args = target_route.presenter_args()

		if hasattr(presenter, action_name) is False:
			raise RuntimeError('No such action "%s" for "%s" presenter' % (action_name, presenter_name))

		action = getattr(presenter, action_name)
		if ismethod(action) is False:
			raise RuntimeError(
				'Unable to execute "%s" action for "%s" presenter' % (action_name, presenter_name)
			)

		args_spec = getfullargspec(action)
		defaults = len(args_spec.defaults) if args_spec.defaults is not None else 0
		action_args = list()
		action_kwargs = dict()

		for i in range(len(args_spec.args)):
			arg = args_spec.args[i]
			if arg == 'self':
				continue

			is_kwarg = i >= (len(args_spec.args) - defaults)

			if is_kwarg is False:
				action_args.append(presenter_args[arg])
			elif arg in presenter_args:
				action_kwargs[arg] = presenter_args[arg]

		return action(*action_args, **action_kwargs)

	@verify_type('paranoid', pattern=str, presenter=(type, str))
	@verify_value('paranoid', presenter=lambda x: issubclass(x, WWebPresenter) or isinstance(x, str))
	def connect(self, pattern, presenter, **kwargs):
		""" Shortcut for self.route_map().connect() method. It is possible to pass presenter class instead of
		its name - in that case such class will be saved in presenter collection and it will be available in
		route matching.

		:param pattern: same as pattern in :meth:`.WWebRouteMap.connect` method
		:param presenter: presenter name or presenter class
		:param kwargs: same as kwargs in :meth:`.WWebRouteMap.connect` method
		:return: None
		"""
		if isinstance(presenter, type) and issubclass(presenter, WWebPresenter) is True:
			self.presenter_collection().add(presenter)
			presenter = presenter.__presenter_name__()
		self.__route_map.connect(pattern, presenter, **kwargs)

	@verify_subclass(presenter=WWebPresenter)
	def add_presenter(self, presenter):
		""" Add presenter to a collection. If the given presenter is a :class:`.WWebEnhancedPresenter` instance
		then public routes are checked (via :meth:`..WWebEnhancedPresenter.__public_routes__` method) and are
		added in this route map

		:param presenter: presenter to add
		:return: None
		"""
		self.__presenter_collection.add(presenter)
		if issubclass(presenter, WWebEnhancedPresenter) is True:
			for route in presenter.__public_routes__():
				self.route_map().append(route)

	@verify_type('paranoid', request=WWebRequestProto, presenter_name=str)
	@verify_type(original_target_route=WWebTargetRoute)
	def proxy(self, request, original_target_route, presenter_name, **kwargs):
		""" Execute the given presenter as a target for the given client request

		:param request: original client request
		:param original_target_route: previous target route
		:param presenter_name: target presenter name
		:param kwargs: presenter arguments
		:return: WWebResponseProto
		"""

		action_kwargs = kwargs.copy()
		action_name = 'index'
		if 'action' in action_kwargs:
			action_name = action_kwargs['action']
			action_kwargs.pop('action')

		original_route = original_target_route.route()
		original_route_map = original_target_route.route_map()
		target_route = WWebTargetRoute(
			presenter_name, action_name, original_route, original_route_map, **action_kwargs
		)
		return self.execute(request, target_route)

	def debugger(self):
		return self.__debugger


class WWebEnhancedPresenter(WWebPresenter, metaclass=ABCMeta):
	""" This is enhanced version of presenter. Besides :class:`.WWebPresenter` class. This class objects
	could use an origin service instance and an origin target route
	"""

	@verify_type('paranoid', request=WWebRequestProto)
	@verify_type(target_route=WWebTargetRoute, service=WWebService)
	def __init__(self, request, target_route, service):
		""" Create new enhanced presenter

		:param request: client request
		:param target_route: source target route
		:param service: source service
		"""
		WWebPresenter.__init__(self, request)
		self.__target_route = target_route
		self.__service = service

	def __target_route__(self):
		""" Return origin target route

		:return: WWebTargetRoute
		"""
		return self.__target_route

	def __service__(self):
		""" Return origin service instance

		:return: WWebService
		"""
		return self.__service

	@classmethod
	def __public_routes__(cls):
		""" Return web routes, that this presenter makes "public" (available to :class:`.WWebService` route map)

		see :meth:`.WWebService.add_presenter` method

		:return: tuple/list of WWebRoute
		"""
		return tuple()

	@verify_type('paranoid', presenter_name=str)
	def __proxy__(self, presenter_name, **kwargs):
		""" Execute the given presenter as a target for the original client request

		:param presenter_name: target presenter name
		:param kwargs: presenter arguments
		:return: WWebResponseProto
		"""
		return self.__service__().proxy(self.__request__(), self.__target_route__(), presenter_name, **kwargs)


class WWebPresenterFactory(WWebPresenterFactoryProto):
	""" WWebPresenterFactoryProto implementation. Can construct classes that do not have their own constructor and
	are derived from WWebPresenter or from WWebEnhancedPresenter classes
	"""

	@staticmethod
	def presenter_constructor(presenter_class, request, target_route, service):
		""" Function that is used for WWebPresenter creating

		:param presenter_class: class to construct
		:param request: original client request
		:param target_route: target route to execute
		:param service: source (parent) service
		:return: WWebPresenter
		"""
		return presenter_class(request)

	@staticmethod
	def enhanced_presenter_constructor(presenter_class, request, target_route, service):
		""" Function that is used for WWebEnhancedPresenter creating

		:param presenter_class: class to construct
		:param request: original client request
		:param target_route: target route to execute
		:param service: source (parent) service
		:return: WWebEnhancedPresenter
		"""
		return presenter_class(request, target_route, service)

	def __init__(self):
		""" Construct new factory
		"""
		WWebPresenterFactoryProto.__init__(self)

		self.__constructors = {
			WWebPresenter.__init__: WWebPresenterFactory.presenter_constructor,
			WWebEnhancedPresenter.__init__: WWebPresenterFactory.enhanced_presenter_constructor
		}

	@verify_subclass(presenter_class=WWebPresenter)
	def _add_constructor(self, presenter_class, constructor_fn):
		self.__constructors[presenter_class.__init__] = constructor_fn

	@verify_subclass(presenter_class=WWebPresenter)
	def instantiable(self, presenter_class):
		""" :meth:`.WWebPresenterFactoryProto.instantiable` method implementation.

		Checks if class doesn't have its own constructor and is derived from WWebPresenter or from
		WWebEnhancedPresenter

		:param presenter_class: class to check
		:return: bool
		"""
		return presenter_class.__init__ in self.__constructors.keys()

	@verify_type('paranoid', request=WWebRequestProto, target_route=WWebTargetRouteProto, service=WWebService)
	@verify_subclass('paranoid', presenter_class=WWebPresenter)
	def instantiate(self, presenter_class, request, target_route, service, *args, **kwargs):
		""" :meth:`.WWebPresenterFactoryProto.instantiate` method implementation.

		Construct new presenter or raise en exception if it isn't possible.

		:param presenter_class: class to construct
		:param request: original client request
		:param target_route: target route to execute
		:param service: source (parent) service
		:param args: additional parameters (do not used)
		:param kwargs: additional parameters (do not used)
		:return: WWebPresenter
		"""
		if self.instantiable(presenter_class) is False:
			raise RuntimeError('Presenter class is not instantiable')
		return self.__constructors[presenter_class.__init__](presenter_class, request, target_route, service)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.crypto.hmac import WHMAC
from wasp_general.crypto.hash import WHash


class TestWHMAC:

	def test(self):
		hmac = WHMAC()
		assert(hmac.digest_generator() == WHash.generator('SHA512'))

		hmac_sha512_result = b'\xc5\xb6\xe3{\x83\xe8A\x1a\xd6\'\x0e\xb1L\x89|a\x8eU\xb2u"\x9e\xefo'
		hmac_sha512_result += b'\xfdkPZ\xd5\x9b\xe2U\xa7\xc3\xc1\x99\xcdf~\xf6\xcbc\xd9\xf8\x9a\xe6\xf9\xe6'
		hmac_sha512_result += b'\xc64_.\xa4\x04\x9e\xac\x9a\x8f\xfe#lfBf'
		assert(hmac.hash(b'11111key2222') == hmac_sha512_result)

		hmac = WHMAC('SHA256')
		assert(hmac.digest_generator() == WHash.generator('SHA256'))
		hmac_sha256_result = b'\x0c\\\xb3\xf1\x08\x84\x18E\xc2l\xf5\xd8\xbc\x85\xd6'
		hmac_sha256_result += b'\x8f?_\xee\xfe\xd9\xf8a\xbc\xa8\x88\xb6o\xd8e\xa3\xf1'
		assert(hmac.hash(b'11111key2222') == hmac_sha256_result)

		hmac = WHMAC.hmac('HMAC-SHA1')
		assert(hmac.digest_generator() == WHash.generator('SHA1'))
		assert(WHMAC.hmac('HMAC-SHA1').digest_generator() == WHMAC.hmac('HMAC_SHA1').digest_generator())
		hmac_sha1_result = b'\xaa\xcb9\x1d\xd4\xeb\xad\xc4s\x03\x83ONN\x8en\xc8\x88I\x1c'
		assert(hmac.hash(b'11111key2222', message=b'some salt') == hmac_sha1_result)

		pytest.raises(ValueError, WHMAC, '???')
